#!/usr/bin/env python3
"""
Module with models for collections of pets and mounts.
"""
from typing import Callable, Dict
from dataclasses import dataclass

from hopla.cli.groupcmds.get_user import HabiticaUser
from hopla.hoplalib.zoo.foodmodels import FeedStatus
from hopla.hoplalib.zoo.petmodels import Mount, Pet, PetMountPair

Zoo = Dict[str, PetMountPair]
"""
Zoo is dictionary with key pet name keys for O(1) access to the
PetMountPair.
"""


@dataclass(frozen=True)
class ZooHelper:
    """Class with helper functions for a Zoo."""
    zoo: Zoo

    def filter_on_pet_mount_pairs(self, predicate: Callable[[PetMountPair], bool]) -> Zoo:
        """Filter the zoo on the pair. This does not change the underlying zoo.

        :param predicate: Include the PetMountPair if the predicate returns True, else
                          omit the pair.
        :return: A filtered Zoo
        """
        return {
            pet_name: pair for (pet_name, pair) in self.zoo.items() if predicate(pair)
        }

    def get_feedable_zoo(self) -> Zoo:
        """Helper function to get only the pets that can be fed in this Zoo."""
        return self.filter_on_pet_mount_pairs(PetMountPair.can_feed_pet)

    def filter_on_pet_name(self, predicate: Callable[[str], bool]) -> Zoo:
        """Filter the zoo on the pet name. This does not change the underlying zoo.

        :param predicate: Include the PetMountPair if the predicate returns True, else
                          omit the pair.
        :return: A filtered zoo.
        """
        return {
            pet_name: pair for (pet_name, pair) in self.zoo.items() if predicate(pet_name)
        }

    def filter_on_pet(self, predicate: Callable[[Pet], bool]) -> Zoo:
        """Filter the zoo on the pet object. This does not change the underlying zoo.

        :param predicate: Include the PetMountPair if the predicate returns True, else
                          omit the pair.
        :return: A filtered zoo.
        """
        return {
            pet_name: pair for (pet_name, pair) in self.zoo.items() if predicate(pair.pet)
        }


class ZooBuilder:
    """
    Class that creates a Zoo from a HabiticaUser using the
    builder design pattern.
    """

    def __init__(self, user: HabiticaUser):
        self.pets: dict = user.get_pets()
        self.mounts: dict = user.get_mounts()
        self.__zoo: Zoo = {}  # empty until build() is called

    def __repr__(self):
        return self.__class__.__name__ + f"({self.__dict__})"

    def build(self) -> Zoo:
        """Create the Zoo. Return the Zoo in case of success"""

        # loop through the pets
        for pet_name, feed_status in self.pets.items():
            pet = Pet(pet_name, feed_status=FeedStatus(feed_status))

            if self.mounts.get(pet_name) is not None:
                mount = Mount(
                    mount_name=pet_name,
                    availability_status=self.mounts[pet_name]
                )
                del self.mounts[pet_name]  # no need to handle the mount twice
            else:
                mount = None
            self.__zoo[pet_name] = PetMountPair(pet=pet, mount=mount)

        # loop through the remaining mounts
        for mount_name, availability_status in self.mounts.items():
            if mount_name not in self.pets:
                # We found a mount without a pet. This is possible for rares.
                mount = Mount(mount_name, availability_status=availability_status)
                self.__zoo[mount_name] = PetMountPair(pet=None, mount=mount)

        return self.__zoo



"""
The module with CLI code that handles the `hopla request` command.
"""
import logging
from typing import Dict, List, Tuple

import click
import requests
from requests import PreparedRequest, Response, Request

from hopla.hoplalib.http import RequestHeaders
from hopla.hoplalib.outputformatter import JsonFormatter

log = logging.getLogger()

BodyParam = Tuple[str, str]


@click.command()
@click.option(
    "--method", "-X", default="GET", show_default=True,
    type=click.Choice(["GET", "POST"]),
    help="The HTTP method."
)
@click.option(
    "--domain", default="https://habitica.com", show_default=True,
    metavar="URL",
    help="The domain name where the Habitica API is hosted."
)
@click.option(
    "--body-param", "body_param_list", multiple=True, type=click.Tuple([str, str]),
    metavar="{KEY} {VALUE}",
    default=[], show_default=False,
    help="A key-value pair for in the JSON body. E.g. --body-param 'dayStart' '0' ."
         "The --body-param option can be used multiple times."
)
@click.option(
    "--show-response-headers", "-I", "show_response_headers_flag",
    is_flag=True, default=False, show_default=True,
    help="Print the HTTP response headers to stdout."
)
@click.option(
    "--show-status-code", "show_status_code_flag",
    is_flag=True, default=False, show_default=True,
    help="Print the HTTP response code (e.g. 201) to stdout."
)
@click.option(
    "--show-response/--no-show-response", "show_response_flag",
    is_flag=True, default=True, show_default=True,
    help="Print the HTTP response body."
)
@click.argument("path")
def request(method: str,
            domain: str,
            body_param_list: List[BodyParam],
            show_response_headers_flag: bool,
            show_status_code_flag: bool,
            show_response_flag: bool,
            path: str):
    # pylint: disable=too-many-arguments
    """Perform a HTTP request on the Habitica API.

    PATH This is the path of the endpoint that you want to perform a HTTP
    request on. For example, /api/v3/groups.

    \b
    Examples
    ----
    # GET: get party information
    $ hopla request /api/v3/groups/party

    \b
    # GET: a user's tasks
    $ hopla request /api/v3/tasks/user

    \b
    # GET: a user's habits
    $ hopla request /api/v3/tasks/user?type=habits

    \b
    # POST: open a mystery item
    $ hopla request --method=POST /api/v3/user/open-mystery-item

    \b
    # POST: send a quest invite for Recidivate pt. 3 to your party
    $ hopla request --method POST /api/v3/groups/party/quests/invite/moonstone3

    \b
    # POST: buy a special item. In this case, a spooky sparkles item.
    $ hopla request -XPOST /api/v3/user/buy-special-spell/spookySparkles

    \b
    # POST: cast the Searing Brightness (Healer) and Tool of Trade (Rogue) spell.
    $ hopla request -XPOST /api/v3/user/class/cast/brightness
    $ hopla request -XPOST /api/v3/user/class/cast/toolsOfTrade

    \b
    # POST: set your custom day start to 1 AM
    $ hopla request -XPOST --body-param dayStart 1 /api/v3/user/custom-day-start


    \b
    # Only show the HTTP response headers
    $ hopla request /api/v3/user --show-response-headers --no-show-response

    \b
    # Only show the HTTP status code
    $ hopla request /api/v3/challenges/groups/party --show-status-code --no-show-response

    \b
    # Show headers, status code and response.
    $ hopla request /api/v3/challenges/groups/party --show-response-headers --show-status-code



    \f
    :return:
    """

    # There is no support for query parameters yet, however, you can use ?y=x in the
    # path to get around this.
    log.debug(f"hopla request {method} {domain=} {path=}\n"
              f"  {body_param_list=}\n"
              f"  {show_response_headers_flag=} {show_status_code_flag=}\n"
              f"  {show_response_flag=}")

    response: Response = perform_request(
        method=method,
        request_endpoint=f"{domain}{path}",
        body_params=dict(body_param_list)
    )

    display_response(
        response=response,
        show_response_headers=show_response_headers_flag,
        show_status_code=show_status_code_flag,
        show_response=show_response_flag,
    )


def perform_request(*, method: str,
                    request_endpoint: str,
                    body_params: Dict[str, str]) -> Response:
    """
    Create and execute an API request according to the specified parameters.

    :param method:
    :param request_endpoint:
    :param body_params:
    :return:
    """
    headers: dict = RequestHeaders().get_default_request_headers()

    http_request = Request(
        method=method,
        url=request_endpoint,
        headers=headers,
        json=body_params or None
    )
    prepared_request: PreparedRequest = http_request.prepare()
    return requests.session().send(prepared_request)


def display_response(response: Response, *,
                     show_response_headers: bool,
                     show_status_code: bool,
                     show_response: bool) -> None:
    """
    Display the requests.Response object in accordance to the specified flags.
    """
    if show_response_headers is True:
        click.echo("RESPONSE HEADERS")
        for header_name, header_value in response.headers.items():
            click.echo(f"{header_name}: {header_value}")

    if show_status_code is True:
        click.echo(f"HTTP Status Code: {response.status_code}")

    if show_response is True:
        # no support for non-JSON output
        click.echo(JsonFormatter(response.json()).format_with_double_quotes())



#!/usr/bin/env python3
"""
Module with throttling logic.
"""
import logging
import time
from dataclasses import dataclass, field
from datetime import datetime, timedelta, timezone
from time import sleep
from typing import Any, Callable, Iterator, List, Optional

import click
from requests import Response
from requests.structures import CaseInsensitiveDict

from hopla.hoplalib.http import ResponseHeaders

log = logging.getLogger()


@dataclass
class ApiRequestThrottler:
    """
    An object that manages a list of api requests.
    """

    def __init__(self, api_requests: List[Callable[[], Any]],
                 *, throttle_seconds: float = 2.5,
                 throttle_limit: int = 25):
        self.api_requests = api_requests
        self.total_calls = len(api_requests)
        self.throttle_seconds = throttle_seconds
        """How long to wait between api calls that are throttled."""
        self.throttle_limit = throttle_limit
        """Maximum of calls that you can do without throttling."""

    def release(self) -> Iterator[Any]:
        """Yields the next api requests ready to be executed.

        :return:
        """
        throttle_enabled: bool = self.exceeds_throttle_limit()

        if throttle_enabled:
            click.echo(f"To prevent overheating Habitica, we'll "
                       f"wait {self.throttle_seconds} seconds every request.",
                       err=True)

        for api_request in self.api_requests:
            yield api_request
            if throttle_enabled:
                self.throttle()

    def execute_all_requests(self) -> None:
        """Execute all API request and discard the return values."""
        for api_request in self.release():
            _ = api_request()

    def throttle(self) -> None:
        """Wait before executing the next api request."""
        sleep(self.throttle_seconds)

    def exceeds_throttle_limit(self, *,
                               api_call_times: Optional[int] = None) -> bool:
        """Return True if we need Hopla to go easy on the Habitica API."""
        call_times: int = api_call_times or self.total_calls
        return call_times > self.throttle_limit


@dataclass
class RateLimitingAwareThrottler:
    """
    Throttler that slows down a list of requests based on HTTP rate-limiting
    headers. This throttler acts as a queue of API requests and its
    `perform_and_yield_response` function acts as a dispatcher.

    [WIKI](https://habitica.fandom.com/wiki/Guidance_for_Comrades#Rate_Limiting)
    """
    api_requests: List[Callable[[], Response]] = field(
        default_factory=list,
        repr=False
    )
    """The queue of API requests. (Currently implemented as a list.)"""
    leeway_seconds: float = 0.25
    """Number of seconds to wait over the requested limit."""
    _is_rate_initialized: bool = field(init=False, default=False)
    """We only know rate limiting information after the first API request."""
    _xrate_limit_remaining: Optional[int] = field(init=False, default=None)
    """
    The number of remaining requests that can be made in the current
    60 second period.

    This value is derived from the X-RateLimit-Remaining header and
    is initialized to None because we won't know it yet at the first
    initialization.
    """

    _xrate_limit_reset: Optional[datetime] = field(init=False, default=None)
    """
    The time on which the current rate-limit will be reset.

    This value derived from the X-RateLimit-Reset header and
    is initialized to None because we won't know it yet at the first
    initialization.
    """

    def __post_init__(self):
        # pylint: disable=pointless-string-statement
        # I think pylint doesn't recognize that these strings act as documentation
        # in the __post_init__ function.

        self._api_requests_remaining: int = len(self.api_requests)
        """The number of Api requests that we haven't executed yet."""

    def perform_and_yield_response(self) -> Iterator[Response]:
        """
        Execute the next request. This function acts as a dispatcher.
        :return:
        """
        for api_request in self.api_requests:
            log.debug(self)
            if self._is_rate_initialized is True and self._throttling_required():
                self._throttle()

            response: Response = api_request()
            self.__update_rate_info(response)

            yield response

    def __update_rate_info(self, response: Response):
        """Use the response to update rate limiting information"""
        self._set_xrate_limit_remaining(response.headers)
        self._set_xrate_limit_reset(response.headers)
        # Consider dequeueing from a queue instead calculating remaining requests.
        self._api_requests_remaining -= 1
        self._is_rate_initialized = True

    def _throttling_required(self) -> bool:
        """Return True when queue is relatively long compared to xrate-limit."""
        return self._api_requests_remaining >= self._xrate_limit_remaining

    def _throttle(self) -> None:
        """Sleep the required time."""
        time.sleep(self._calculate_sleep_time())

    def _calculate_sleep_time(self) -> float:
        """
        Calculate how many seconds to sleep to evenly spread out the
        API requests over the remaining xrate reset time.
        """
        now_utc: datetime = datetime.now(timezone.utc)
        time_till_reset: timedelta = self._xrate_limit_reset - now_utc

        if self._xrate_limit_remaining == 0:
            # No request remains. Let's chill out a lot.
            sleep_time: float = time_till_reset.total_seconds() + 5.
        elif self._xrate_limit_remaining == 1:
            # Only 1 request remains. Let's chill out.
            sleep_time: float = time_till_reset.total_seconds() + 3
        elif self._xrate_limit_remaining == 2:
            # Only 2 requests remain. By slowing down in this manner, we can run
            # hopla commands concurrently.
            sleep_time: float = time_till_reset.total_seconds() + 2
        else:
            # Go fast.
            time_interval: timedelta = time_till_reset / self._xrate_limit_remaining
            sleep_time: float = time_interval.total_seconds()

        # use max(..., 0) for when x-rate limit is in the past.
        return max(sleep_time, 0) + self.leeway_seconds

    def _set_xrate_limit_remaining(self, headers: CaseInsensitiveDict):
        """Set the xrate_limit_remaining from a given Habitica API response headers."""
        calls_remaining: str = headers[ResponseHeaders.XRATE_LIMIT_REMAINING_HEADER_NAME]
        self._xrate_limit_remaining = int(calls_remaining)

    def _set_xrate_limit_reset(self, headers: CaseInsensitiveDict):
        """Set the xrate_limit_reset from a given Habitica API response headers."""
        reset_datetime_str: str = headers[ResponseHeaders.XRATE_LIMIT_RESET_HEADER_NAME]
        self._xrate_limit_reset = datetime.strptime(
            reset_datetime_str.split(" (")[0],  # oef... painful
            "%a %b %d %Y %H:%M:%S %Z%z"
            # example:
            # "Mon Oct 16 2022 13:49:39 GMT+0000 (Coordinated Universal Time)",
        )



"""
A modules with algorithms for feeding multiple pets at once.
"""
from dataclasses import dataclass
from typing import Iterator, List
from copy import deepcopy

from hopla.hoplalib.zoo.foodmodels import FoodStockpile
from hopla.hoplalib.zoo.petmodels import Pet
from hopla.hoplalib.zoo.zoomodels import Zoo, ZooHelper


@dataclass(frozen=True)
class FeedPlanItem:
    """An item of a feed plan.

    An item on the feed plan describes exactly what food to feed 1 single pet.
    """
    pet_name: str
    food_name: str
    times: int

    def format_item(self) -> str:
        """A to string function for nice terminal output"""
        return f"Pet {self.pet_name} will get {self.times} {self.food_name}."


@dataclass
class FeedPlan:
    """A plan to feed pets.

    This class acts as a wrapper around a list with feeding parameters.
    """

    def __init__(self):
        self.__feed_plan: [FeedPlanItem] = []

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(__feed_plan={self.__feed_plan})"

    def __iter__(self) -> Iterator[FeedPlanItem]:
        return iter(self.__feed_plan)

    def __len__(self) -> int:
        """Return the number of FeedPlanItems in this feed plan."""
        return len(self.__feed_plan)

    def isempty(self) -> bool:
        """Return true if the feed plan is empty."""
        return len(self) == 0

    def add_to_feed_plan(self, *, pet_name: str, food_name: str, times: int) -> None:
        """Add an item to the feed plan.

        :param pet_name: name of the pet to feed
        :param food_name: name of the food to give
        :param times: number of food items to give
        :return: None
        """
        feed_item = FeedPlanItem(
            pet_name=pet_name,
            food_name=food_name,
            times=times
        )
        self.__feed_plan.append(feed_item)

    @property
    def feed_plan(self) -> List[FeedPlanItem]:
        """Return the feed plan.

        :return: The feed plan. This will be empty when
        no feed plan items have been added yet.
        """
        return self.__feed_plan

    def format_plan(self) -> str:
        """A to string function for nice terminal output"""
        return "\n".join([item.format_item() for item in self.__feed_plan])


@dataclass
class FeedAlgorithm:
    """
    This class contains an algorithm that makes a plan to distribute food
    from a food stockpile over all pets.
    """

    def __init__(self, *, zoo: Zoo, stockpile: FoodStockpile):
        self.__zoo: Zoo = ZooHelper(deepcopy(zoo)).get_feedable_zoo()
        self.__stockpile = deepcopy(stockpile)

        self.__feed_plan = FeedPlan()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(\n"
            f"  __feed_plan={self.__feed_plan.format_plan()}\n"
            f")"
        )

    @property
    def stockpile(self) -> FoodStockpile:
        """Return the stockpile."""
        return self.__stockpile

    @property
    def feed_plan(self) -> FeedPlan:
        """Return the underlying feed plan.

        This plan will be empty before calling make_plan.
        """
        return self.__feed_plan

    def make_plan(self) -> FeedPlan:
        """Make the plan.

        This function removes food from the stockpile and adds feed items
        to the feed plan.
        """
        helper = ZooHelper(self.__zoo)
        gen1_zoo: Zoo = helper.filter_on_pet(Pet.is_generation1_pet)
        quest_zoo: Zoo = helper.filter_on_pet(Pet.is_quest_pet)
        magic_zoo: Zoo = helper.filter_on_pet(Pet.is_magic_hatch_pet)

        self.__make_plan(gen1_zoo)
        self.__make_plan(quest_zoo)
        self.__make_plan(magic_zoo)

        return self.__feed_plan

    def __make_plan(self, zoo: Zoo):
        """Make plan for the specified zoo.

        This function assumes that only feedable pets are passed.
        """
        for pet_name, pair in zoo.items():
            pet: Pet = pair.pet

            if pet.has_just_1_favorite_food():
                food_name: str = pet.favorite_food()
            else:
                food_name: str = self.__stockpile.get_most_abundant_food()

            times: int = pet.required_food_items_until_mount(food_name)
            if self.__stockpile.has_sufficient(food_name, n=times):
                subtract_times: int = -times
                self.__stockpile.add_food(food_name,
                                          n=subtract_times)
                self.__feed_plan.add_to_feed_plan(
                    pet_name=pet_name,
                    food_name=food_name,
                    times=times
                )



from hopla.cli.get_user.info import filter_on_user_info_name
from hopla.cli.groupcmds.get_user import HabiticaUser


class TestPrefilterOnUserInfo:
    def test_all_returns_user_unchanged(self):
        test_user: HabiticaUser = get_test_user()

        filter_result = filter_on_user_info_name(test_user, "all")

        assert filter_result == test_user.user_dict

    def test_gems_return_gems(self):
        balance = 2
        test_user = HabiticaUser({"balance": balance})

        filter_result = filter_on_user_info_name(test_user, "gems")

        expected_gems = balance * 4
        assert filter_result == expected_gems
        assert type(expected_gems) is int

    def test_id_returns_id(self):
        test_user: HabiticaUser = get_test_user()

        filter_result = filter_on_user_info_name(test_user, "id")

        expected_id = test_user["id"]
        assert filter_result == expected_id
        assert type(filter_result) is str

    def test_guilds_returns_list(self):
        test_user: HabiticaUser = get_test_user()

        filter_result = filter_on_user_info_name(test_user, "guilds")

        expected_guilds = test_user["guilds"]
        assert filter_result == expected_guilds
        assert isinstance(filter_result, list)

    def test_tags_returns_list_of_objects(self):
        test_user: HabiticaUser = get_test_user()

        filter_result = filter_on_user_info_name(test_user, "tags")

        expected_tags = test_user["tags"]
        assert filter_result == expected_tags
        assert isinstance(filter_result, list)


def get_test_user() -> HabiticaUser:
    """
    This user dict has been edited as follows:
    * All ids have been changed completely, only their syntax is intact
    * All dates have been changed (most to the future)
    * long lists have been cut down significantly to remove the duplication
      of data that is largely similar in syntax (.g. finished quests).
      The was to keep the original range of data intact (e.g. at least
      keep a large value, negative value, and default value for feed
      status of pets. e.g. within 'history' dates came in multiple formats,
      (millis and iso dates) so both formats were kept).
    * care was taken to not modify the underlying structure and
      realisticness of the data. However, the data is note entirely valid
      because e.g. available eggs don't match finished quests. Sort order
      ids dont match actual ids or do not exist, etc.
    * null from javascript were replaced with "null" strings
    * unicode chars were remove (\\u)
    """
    contributions: str = (
        "[x xx GitHub](https://github.com/x), "
        "[xxx on the wiki](https://habitica.fandom.com/wiki/User:xxx)\n"
        "* Tier 2 (in progress): "
        "[dsklaf DSL dklasfalsjfjsts](https://github.com/xxx/xxxx/pull/21389)"
    )
    semi_realistic_user: dict = {
        "auth": {
            "local": {
                "username": "vioman",
                "lowerCaseUsername": "vioman",
                "email": "vioman@vioman.io"
            },
            "timestamps": {
                "created": "2024-03-20T20:23:38.119Z",
                "loggedin": "2027-07-08T04:47:51.661Z",
                "updated": "2028-07-08T05:34:02.781Z"
            },
            "facebook": {},
            "google": {},
            "apple": {}
        },
        "achievements": {
            "ultimateGearSets": {
                "healer": False,
                "wizard": True,
                "rogue": True,
                "warrior": False
            },
            "streak": 172,
            "challenges": [],
            "perfect": 109,
            "quests": {
                "goldenknight2": 2,
                "goldenknight3": 1,
                "unicorn": 3,
                "egg": 1,
                "slime": 1,
                "rooster": 1
            },
            "completedTask": True,
            "createdTask": True,
            "hatchedPet": True,
            "fedPet": True,
            "purchasedEquipment": True,
            "joinedGuild": True,
            "partyUp": True,
            "partyOn": True,
            "joinedChallenge": True,
            "freshwaterFriends": True,
            "dustDevil": True,
            "boneCollector": True,
            "backToBasics": True,
            "monsterMagus": True,
            "thankyou": 1,
            "goodAsGold": True,
            "primedForPainting": True,
            "tickledPink": True,
            "congrats": 1,
            "beastMaster": True,
            "seeingRed": True,
            "violetsAreBlue": True,
            "rosyOutlook": True,
            "skeletonCrew": True,
            "allYourBase": True,
            "undeadUndertaker": True,
            "aridAuthority": True,
            "allThatGlitters": True,
            "pearlyPro": True,
            "wildBlueYonder": True,
            "mountMaster": True,
            "redLetterDay": True,
            "triadBingo": True,
            "seafoam": 5,
            "rebirthLevel": 100,
            "rebirths": 1,
            "justAddWater": True,
            "habiticaDays": 1,
            "bugBonanza": True,
            "bareNecessities": True,
            "mindOverMatter": True
        },
        "backer": {},
        "contributor": {
            "contributions": contributions,
            "level": 1,
            "text": "Comrade, Blacksmith",
            "admin": False
        },
        "purchased": {
            "ads": False,
            "txnCount": 1,
            "skin": {},
            "hair": {},
            "shirt": {},
            "background": {
                "violet": True,
                "blue": True,
                "green": True,
                "purple": True,
                "red": True,
                "yellow": True,
                "blacksmithy": True,
                "crystal_cave": True,
                "distant_castle": True
            },
            "plan": {
                "consecutive": {
                    "count": 2,
                    "offset": 1,
                    "gemCapExtra": 5,
                    "trinkets": 1
                },
                "quantity": 1,
                "extraMonths": 0,
                "gemsBought": 9000,
                "mysteryItems": [],
                "dateUpdated": "2034-06-01T04:52:21.012Z",
                "customerId": "X-XX111AA3",
                "dateCreated": "2034-05-03T04:52:04.117Z",
                "dateTerminated": "null",
                "lastReminderDate": "null",
                "owner": "0000fd00-ffe0-0fb0-b0fa-0d00b0000ff0",
                "paymentMethod": "something",
                "planId": "basic_6mo"
            }
        },
        "flags": {
            "tour": {
                "intro": -2,
                "classes": -1,
                "stats": -1,
                "tavern": -2,
                "party": -2,
                "guilds": -2,
                "challenges": -1,
                "market": -2,
                "pets": -2,
                "mounts": -1,
                "hall": -2,
                "equipment": -2
            },
            "tutorial": {
                "common": {
                    "habits": True,
                    "dailies": True,
                    "todos": True,
                    "rewards": True,
                    "party": True,
                    "pets": True,
                    "gems": True,
                    "skills": True,
                    "classes": True,
                    "tavern": True,
                    "equipment": True,
                    "items": True,
                    "mounts": True,
                    "inbox": True,
                    "stats": True
                },
                "ios": {
                    "addTask": False,
                    "editTask": False,
                    "deleteTask": False,
                    "filterTask": False,
                    "groupPets": False,
                    "inviteParty": False,
                    "reorderTask": False
                }
            },
            "customizationsNotification": True,
            "showTour": True,
            "dropsEnabled": False,
            "itemsEnabled": True,
            "lastNewStuffRead": "aaaaaaaa-bbbb-ccccccccc-dddddddddddd",
            "rewrite": True,
            "classSelected": True,
            "rebirthEnabled": True,
            "recaptureEmailsPhase": 0,
            "weeklyRecapEmailsPhase": 0,
            "communityGuidelinesAccepted": True,
            "cronCount": 171,
            "welcomed": True,
            "armoireEnabled": True,
            "armoireOpened": True,
            "armoireEmpty": False,
            "cardReceived": False,
            "warnedLowHealth": True,
            "verifiedUsername": True,
            "levelDrops": {
                "atom1": True,
                "vice1": True,
                "goldenknight1": True,
                "moonstone1": True
            },
            "lastWeeklyRecap": "2022-03-21T20:23:38.119Z",
            "onboardingEmailsPhase": "9-b-1111111112234",
            "lastFreeRebirth": "2022-07-15T18:42:05.630Z",
            "newStuff": False
        },
        "history": {
            "exp": [
                {
                    "date": 1616475808219,
                    "value": 708.66015625
                },
                {
                    "date": 1617253819642,
                    "value": 10388.243710296229
                },
                {
                    "date": 1619843534792,
                    "value": 13132.75
                },
                {
                    "date": "2021-05-04T04:42:22.770Z",
                    "value": 14551
                }
            ],
            "todos": [
                {
                    "date": 1616475808219,
                    "value": -12.781333149500004
                },
                {
                    "date": 1617253819642,
                    "value": -219.47350071428951
                },
                {
                    "date": "2021-06-26T07:04:53.202Z",
                    "value": -1951.1495238060916
                },
                {
                    "date": "2021-06-27T04:29:01.504Z",
                    "value": -2065.735204333207
                },
                {
                    "date": "2021-09-08T04:44:44.422Z",
                    "value": -14520.782668485932
                },
                {
                    "date": "2021-09-09T04:47:51.661Z",
                    "value": -14927.3409270358
                }
            ]
        },
        "items": {
            "gear": {
                "equipped": {
                    "armor": "armor_rogue_5",
                    "head": "head_special_pageHelm",
                    "shield": "shield_special_goldenknight",
                    "headAccessory": "headAccessory_armoire_gogglesOfBookbinding",
                    "weapon": "weapon_special_2",
                    "eyewear": "eyewear_armoire_goofyGlasses",
                    "body": "body_armoire_cozyScarf",
                    "back": "back_special_aetherCloak"
                },
                "costume": {
                    "armor": "armor_armoire_autumnEnchantersCloak",
                    "head": "head_wizard_3",
                    "shield": "shield_base_0",
                    "weapon": "weapon_base_0",
                    "headAccessory": "headAccessory_base_0",
                    "eyewear": "eyewear_base_0",
                    "back": "back_base_0",
                    "body": "body_base_0"
                },
                "owned": {
                    "headAccessory_special_blackHeadband": True,
                    "headAccessory_special_blueHeadband": True,
                    "back_mystery_202109": True,
                    "headAccessory_mystery_202109": True,
                    "head_special_snowSovereignCrown": True,
                    "armor_special_snowSovereignRobes": True,
                    "weapon_armoire_heraldsBuisine": True,
                    "armor_armoire_heraldsTunic": True,
                    "head_armoire_heraldsCap": True
                }
            },
            "special": {
                "snowball": 0,
                "spookySparkles": 0,
                "shinySeed": 5,
                "seafoam": 2,
                "valentine": 0,
                "valentineReceived": [],
                "nye": 0,
                "nyeReceived": [],
                "greeting": 0,
                "greetingReceived": [],
                "thankyou": 0,
                "thankyouReceived": [],
                "birthday": 0,
                "birthdayReceived": [],
                "congrats": 0,
                "congratsReceived": [],
                "getwell": 0,
                "getwellReceived": [],
                "goodluck": 0,
                "goodluckReceived": []
            },
            "lastDrop": {
                "count": 11,
                "date": "2044-09-09T04:52:33.115Z"
            },
            "pets": {
                "Dragon-Desert": 5,
                "Triceratops-CottonCandyBlue": -1,
                "Wolf-Windup": 5,
                "Beetle-Shade": -1,
                "Beetle-CottonCandyPink": 15,
                "Beetle-Skeleton": -1,
                "Snail-Desert": -1
            },
            "eggs": {
                "Dragon": 188,
                "LionCub": 196,
                "Axolotl": 0,
                "Seahorse": 3,
                "Whale": 3,
                "Beetle": 0,
                "Snail": 2,
                "Rock": 3,
                "Egg": 10,
                "Slime": 3,
                "Rooster": 3
            },
            "hatchingPotions": {
                "CottonCandyBlue": 129,
                "RoyalPurple": 3,
                "Golden": 112,
                "CottonCandyPink": 138,
                "Bronze": 0,
                "BlackPearl": 0,
                "Windup": 0
            },
            "food": {
                "RottenMeat": 406,
                "Potatoe": 362,
                "Saddle": 1,
                "Cake_White": 1,
                "Cake_Zombie": 1
            },
            "mounts": {
                "Fox-RoyalPurple": True,
                "FlyingPig-Zombie": True,
                "Fox-Base": True,
                "Fox-Golden": True,
                "Turtle-CottonCandyPink": True,
                "Whale-White": True,
                "Seahorse-Red": True,
                "Seahorse-CottonCandyPink": True,
                "Axolotl-Base": True,
                "Nudibranch-CottonCandyBlue": True,
                "Triceratops-Base": True,
                "SeaSerpent-Base": True,
                "Hippo-Base": True,
                "TigerCub-RoyalPurple": True,
                "PandaCub-RoyalPurple": True,
                "Dolphin-Base": True,
                "Gryphon-RoyalPurple": True,
                "Beetle-Shade": True,
                "Beetle-Skeleton": True,
                "Snail-Desert": True
            },
            "quests": {
                "atom1": 2,
                "basilist": 1,
                "moon3": 1,
                "goldenknight1": 2,
                "vice2": 0,
                "vice3": 0
            },
            "currentPet": "Wolf-Base",
            "currentMount": "Aether-Invisible"
        },
        "invitations": {
            "guilds": [],
            "party": {},
            "parties": []
        },
        "party": {
            "quest": {
                "progress": {
                    "up": 22.051118316159503,
                    "down": 0,
                    "collectedItems": 14,
                    "collect": {}
                },
                "RSVPNeeded": False,
                "key": "solarSystem",
                "completed": "null"
            },
            "order": "auth.timestamps.loggedin",
            "orderAscending": "asc",
            "_id": "eeeeeeee-bbbb-0000-dddd-afedfeafedad"
        },
        "preferences": {
            "hair": {
                "color": "blond",
                "base": 3,
                "bangs": 2,
                "beard": 0,
                "mustache": 0,
                "flower": 6
            },
            "emailNotifications": {
                "unsubscribeFromAll": True,
                "newPM": True,
                "kickedGroup": True,
                "wonChallenge": True,
                "giftedGems": True,
                "giftedSubscription": True,
                "invitedParty": True,
                "invitedGuild": True,
                "questStarted": True,
                "invitedQuest": True,
                "importantAnnouncements": True,
                "weeklyRecaps": True,
                "onboarding": True,
                "majorUpdates": True,
                "subscriptionReminders": False
            },
            "pushNotifications": {
                "unsubscribeFromAll": False,
                "newPM": False,
                "wonChallenge": True,
                "giftedGems": False,
                "giftedSubscription": False,
                "invitedParty": False,
                "invitedGuild": False,
                "questStarted": False,
                "invitedQuest": False,
                "majorUpdates": False,
                "mentionParty": False,
                "mentionJoinedGuild": False,
                "mentionUnjoinedGuild": False,
                "partyActivity": False
            },
            "suppressModals": {
                "levelUp": False,
                "hatchPet": False,
                "raisePet": True,
                "streak": False
            },
            "tasks": {
                "groupByChallenge": False,
                "confirmScoreNotes": False
            },
            "dayStart": 0,
            "size": "slim",
            "hideHeader": False,
            "skin": "f5a76e",
            "shirt": "pink",
            "timezoneOffset": -120,
            "sound": "danielTheBard",
            "chair": "none",
            "allocationMode": "taskbased",
            "autoEquip": False,
            "dateFormat": "dd/MM/yyyy",
            "sleep": False,
            "stickyHeader": True,
            "disableClasses": False,
            "newTaskEdit": False,
            "dailyDueDefaultView": True,
            "advancedCollapsed": True,
            "toolbarCollapsed": False,
            "reverseChatOrder": False,
            "displayInviteToPartyWhenPartyIs1": True,
            "improvementCategories": [],
            "language": "en",
            "webhooks": {},
            "background": "distant_castle",
            "timezoneOffsetAtLastCron": -120,
            "costume": True,
            "automaticAllocation": False
        },
        "profile": {
            "name": "vioman",
            "imageUrl": "https://example.com/my.jpg",
            "blurb": "Creator of Hopla"
        },
        "stats": {
            "buffs": {
                "str": 50,
                "int": 50,
                "per": 50,
                "con": 50,
                "stealth": 0,
                "streaks": False,
                "snowball": False,
                "spookySparkles": False,
                "shinySeed": False,
                "seafoam": False
            },
            "training": {
                "int": 0,
                "per": 0,
                "str": 0,
                "con": 0
            },
            "hp": 47.10095169004215,
            "mp": 11.912692307692595,
            "exp": 701,
            "gp": 523808.2384303317,
            "lvl": 101,
            "class": "wizard",
            "points": 0,
            "str": 0,
            "con": 0,
            "int": 12,
            "per": 88,
            "toNextLevel": 3700,
            "maxHealth": 50,
            "maxMP": 254
        },
        "inbox": {
            "newMessages": 0,
            "optOut": False,
            "blocks": [
                "31aa247a-5432-0000-b6b3-111111e20eb0"
            ],
            "messages": {
                "aaaaaa82-b9b7-ccc4-dddd-11111148f725": {
                    "sent": True,
                    "flagCount": 0,
                    "_id": "aaaaaa82-b9b7-ccc4-dddd-11111148f725",
                    "ownerId": "79551d98-31e9-42b4-b7fa-9d89b0944319",
                    "flags": {},
                    "id": "aaaaaaaa-d9b7-42a4-0000-00000048f725",
                    "text": "dsjk :blush:. \nR .asdjfkl asdflasf'asdfljksdf jsdalf   :smiley:",
                    "unformattedText": "sdaf :blush:. fjkl.\nI  djsklajw :smiley:",
                    "info": {},
                    "timestamp": "2024-03-09T04:54:32.565Z",
                    "likes": {},
                    "uuid": "aaaaaaaa-324a-aaaa-abbb-bbbbbbbbbbeb",
                    "contributor": {
                        "contributions": contributions,
                        "level": 1,
                        "text": "Comrade, Blacksmith",
                        "admin": False
                    },

                    "backer": {},
                    "user": "minsun",
                    "username": "minsun",
                    "userStyles": {
                        "items": {
                            "gear": {
                                "costume": {
                                    "armor": "armor_armoire_autumnEnchantersCloak",
                                    "head": "head_wizard_3",
                                    "shield": "shield_base_0",
                                    "weapon": "weapon_base_0",
                                    "headAccessory": "headAccessory_base_0",
                                    "eyewear": "eyewear_base_0",
                                    "back": "back_base_0",
                                    "body": "body_base_0"
                                },
                                "equipped": {
                                    "armor": "armor_rogue_5",
                                    "head": "head_special_pageHelm",
                                    "shield": "shield_special_goldenknight",
                                    "headAccessory": "headAccessory_armoire_gogglesOfBookbinding",
                                    "weapon": "weapon_special_2",
                                    "eyewear": "eyewear_armoire_goofyGlasses",
                                    "body": "body_armoire_cozyScarf",
                                    "back": "back_special_aetherCloak"
                                }
                            },
                            "currentMount": "Aether-Invisible",
                            "currentPet": "Wolf-Base"
                        },
                        "preferences": {
                            "hair": {
                                "color": "blond",
                                "base": 3,
                                "bangs": 2,
                                "beard": 0,
                                "mustache": 0,
                                "flower": 6
                            },
                            "skin": "f5a76e",
                            "shirt": "pink",
                            "chair": "none",
                            "size": "slim",
                            "background": "distant_castle",
                            "costume": True
                        },
                        "stats": {
                            "class": "wizard",
                            "buffs": {
                                "seafoam": False,
                                "shinySeed": False,
                                "spookySparkles": False,
                                "snowball": False
                            }
                        }
                    }
                }
            }
        },
        "tasksOrder": {
            "habits": [
                "dddddddd-30bc-dddd-bebb-fd713096ac83",
                "b09897da-3249-4131-862a-fffffffff765"
            ],
            "dailys": [
                "3f70cf50-bbbb-42f5-a279-aaaaaaaaaaa8",
                "140791fb-b6be-cccc-9918-aaaaaaaaaaa1"
            ],
            "todos": [
                "aaaaaaaa-02ef-4f8e-91b8-b78d03fffba5",
                "bbbbbb7c-7f96-4672-b0ad-85422fff5cf0"
            ],
            "rewards": []
        },
        "_v": 54081,
        "balance": 24.5,
        "challenges": [
            "10101010-1000-4fe1-b6a4-84ce8ef8885d"
        ],
        "guilds": [
            "10101011-1010-4e52-87de-5c040ed0f8f8",
            "fffffff0-087f-4b08-9432-86fe77c2f1ef"
        ],
        "loginIncentives": 177,
        "invitesSent": 0,
        "pinnedItemsOrder": [],
        "_id": "aaaaaaa8-aaa9-42b4-b7fa-9d89b0944319",
        "lastCron": "2026-09-09T04:47:51.661Z",
        "newMessages": {
            "aaaaaaaa-ffff-aaa9-874e-e6e14c53549e": {
                "name": "asdjfk RJKLE dfjklsa LJRK",
                "value": True
            },
            "aaaaaaa0-087f-4b08-9432-86fe77c2f1ef": {
                "name": "dsklfa jkldsa fadskjlfnts and health professionals",
                "value": True
            }
        },
        "notifications": [
            {
                "type": "NEW_CHAT_MESSAGE",
                "data": {
                    "group": {
                        "id": "aaaaaaaa-1fae-4b39-111e-da1fa1df549e",
                        "name": "dsjkladsaf kljafes of Habitica"
                    }
                },
                "id": "aaaa370d-0000-000b-0003-1111166ecd93",
                "seen": False
            },
            {
                "type": "NEW_CHAT_MESSAGE",
                "data": {
                    "group": {
                        "id": "aaaaaaaa-087f-4b08-9432-86fe77c2f1ef",
                        "name": "dsaklf,sd sdjkla asdf, kdlasfjkdl"
                    }
                },
                "id": "aaaaae61-a31b-4588-a54a-23d51c42dbc8",
                "seen": False
            }
        ],
        "tags": [
            {
                "name": "Work",
                "id": "dddddddd-30bc-dddd-bebb-fd713096ac83"
            },
            {
                "id": "b09897da-3249-4131-862a-fffffffff765",
                "name": "BXX2024",
                "challenge": True
            }
        ],
        "extra": {},
        "pushDevices": [
            {
                "regId": "aaaaaaaaaaaaaaa:PPPPPPPPPPPPp099999999999_dsklaffjasdkl-AAAA",
                "type": "android",
                "createdAt": "2023-02-25T16:19:50.079Z",
                "updatedAt": "2025-09-09T04:54:32.599Z"
            }
        ],
        "webhooks": [],
        "pinnedItems": [
            {
                "type": "potion",
                "path": "potion"
            },
            {
                "type": "armoire",
                "path": "armoire"
            },
            {
                "path": "special.gems",
                "type": "gem"
            }
        ],
        "unpinnedItems": [
            {
                "path": "gear.flat.weapon_special_spring2021Rogue",
                "type": "marketGear"
            },
            {
                "path": "gear.flat.armor_special_spring2021Rogue",
                "type": "marketGear"
            },
            {
                "path": "gear.flat.head_special_spring2021Rogue",
                "type": "marketGear"
            },
            {
                "path": "gear.flat.shield_special_spring2021Rogue",
                "type": "marketGear"
            }
        ],
        "_ABTests": {
            "onboardingPushNotification": "Onboarding-Step6-Phasea-VersionD"
        },
        "migration": "20220924_pet_group_achievements",
        "id": "79553d98-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
        "needsCron": False

    }
    return HabiticaUser(semi_realistic_user)



#!/usr/bin/env python3
"""
Module with algorithms for hatching many eggs.
"""
from dataclasses import dataclass, field
from typing import Any, List

from hopla.hoplalib.hatchery.eggmodels import Egg, EggCollection, EggException
from hopla.hoplalib.hatchery.hatchpotionmodels import HatchPotion, \
    HatchPotionCollection
from hopla.hoplalib.zoo.petmodels import Pet


@dataclass(frozen=True)
class HatchPlanItem:
    """Plan to hatch a specific egg with a specified potion."""
    egg: Egg
    potion: HatchPotion

    def __post_init__(self):
        if self.egg.can_be_hatched_by(potion=self.potion) is False:
            msg = f"Cannot hatch {self.egg.name} with {self.potion.name} potion"
            raise EggException(msg)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, HatchPlanItem) is False:
            return False
        # We don't care about quantity here. We always use 1 egg and 1 potion
        # for any hatching in our hatch plan.
        return other.egg.name == self.egg.name and other.potion.name == self.potion.name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.egg.name}: {self.potion.name})"

    def result_pet_name(self) -> str:
        """
        Return the pet name that comes from hatching the egg with the given potion.
        """
        return f"{self.egg.name}-{self.potion.name}"

    def format_item(self) -> str:
        """Terminal-printable representation of this HatchPlanItem."""
        return (
            f"A {self.egg.name} egg will be hatched by a {self.potion.name} potion.\n"
        )

    def can_hatch_with_pet(self, pet: Pet) -> bool:
        """
        Return True if we can hatch this item without clashing with the given pet.
        If the hatching would clash, return False

        :param pet: The pet to check against.
        :return: True if we can safely hatch if we only had this pet.
        """
        return (pet.is_available() is False) or (pet.name != self.result_pet_name())

    def can_hatch_with_pets(self, pets: List[Pet]) -> bool:
        """
        Return True if we can hatch this item without clashing with the given pets.
        If any of the hatchings would clash, return False

        :param pets: The pets to check against.
        :return: True if we can safely hatch even if we had all these pets.
        """
        return all(self.can_hatch_with_pet(pet) for pet in pets)


@dataclass
class HatchPlan:
    """Plan to hatch a specific eggs with specified potions."""

    plan: List[HatchPlanItem] = field(init=False, default_factory=list)

    def __iter__(self):
        return iter(self.plan)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.plan if len(self.plan) > 0 else 'empty'})"

    def __len__(self) -> int:
        return len(self.plan)

    def is_empty(self) -> bool:
        """Return true if there no items in the hatch plan."""
        return len(self.plan) == 0

    def format_plan(self) -> str:
        """Terminal-printable representation of this HatchPlan."""
        return "".join([item.format_item() for item in self.plan])

    def add(self, *, egg: Egg, potion: HatchPotion) -> "HatchPlan":
        """Make the plan to hatch the egg with the specified potion.

        :param egg:
        :param potion:
        :return: self for chaining
        """
        hatch_plan_item = HatchPlanItem(egg=egg, potion=potion)
        self.plan.append(hatch_plan_item)
        return self

    def remove_hatch_item_if_pet_available(self, pets: List[Pet]) -> "HatchPlan":
        """Remove the hatching plan items for pets that are available.

        :param pets: list of pets to check for clashes with the hatching
        :return: self for chaining
        """
        pet_names: List[str] = [pet.name for pet in pets if pet.is_available()]

        self.plan = [hatch_item for hatch_item in self.plan
                     if hatch_item.result_pet_name() not in pet_names]

        return self


@dataclass
class HatchPlanMaker:
    """
    Object that takes a collection of eggs and hatching potions and
    return a plan to hatch the eggs.
    """

    def __init__(self, *,
                 egg_collection: EggCollection,
                 hatch_potion_collection: HatchPotionCollection,
                 pets: List[Pet]):
        self.__eggs = egg_collection
        self.__potions = hatch_potion_collection
        self.__pets = pets
        self.__hatch_plan = HatchPlan()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(\n"
            f"    {self.__eggs},\n"
            f"    {self.__potions},\n"
            f"    plan={self.__hatch_plan}\n"
            ")"
        )

    def make_plan(self) -> HatchPlan:
        """Given an collection of eggs and hatching potions, make a plan to hatch the eggs."""
        for egg in self.__eggs.values():
            for potion in self.__potions.values():
                can_hatch: bool = (
                        egg.can_be_hatched_by(potion)
                        and HatchPlanItem(egg, potion).can_hatch_with_pets(self.__pets)
                )
                if can_hatch:
                    self.__hatch_plan.add(egg=egg, potion=potion)
                    self.__eggs.remove_egg(egg)
                    self.__potions.remove_hatch_potion(potion)

        return self.__hatch_plan



#!/usr/bin/env python3
"""
The module with CLI code that handles the `hopla hatch` GROUP command.
"""

import sys
from typing import NoReturn

import click
import requests

from hopla.hoplalib.hatchery.hatchcontroller import HatchRequester


@click.group()
def hatch():
    """GROUP for hatching eggs."""


def hatch_egg(*, egg_name: str, potion_name: str) -> NoReturn:
    """
    Hatch an egg by performing an API request and echo the result to the
    terminal.
    """
    requester = HatchRequester(
        egg_name=egg_name,
        hatch_potion_name=potion_name
    )
    response: requests.Response = requester.post_hatch_egg_request()
    json: dict = response.json()
    if json["success"] is True:
        click.echo(f"Successfully hatched a {egg_name}-{potion_name}.")
        sys.exit(0)

    click.echo(f"{json['error']}: {json['message']}")
    sys.exit(1)



"""
The module with CLI code that handles the `hopla feed` command.
"""
import logging
import sys
from dataclasses import dataclass
from typing import NoReturn, Optional, Union

import click
import requests

from hopla.cli.groupcmds.get_user import HabiticaUser, HabiticaUserRequest
from hopla.hoplalib.errors import YouFoundABugRewardError
from hopla.hoplalib.zoo.feed_clickhelper import get_feed_data_or_exit
from hopla.hoplalib.zoo.fooddata import FoodData
from hopla.hoplalib.zoo.foodmodels import FoodStockpile, FoodStockpileBuilder
from hopla.hoplalib.zoo.petcontroller import FeedPostRequester
from hopla.hoplalib.zoo.petdata import PetData
from hopla.hoplalib.zoo.petmodels import Pet, PetMountPair
from hopla.hoplalib.zoo.zoomodels import Zoo, ZooBuilder

log = logging.getLogger()

MIN_FEED_TIMES = 0
MAX_FEED_TIMES = 23
valid_feed_amount_range = click.IntRange(min=MIN_FEED_TIMES,
                                         max=MAX_FEED_TIMES,
                                         clamp=True)
"""pets can eat up to 23 units of non-preferred foods"""


def print_favorite_food_and_exit(pet_name: str):
    """Print favorite food for the specified pet and always exit."""
    pet = Pet(pet_name)
    click.echo(pet.favorite_food())
    sys.exit()


def get_feed_times_until_mount(pet_name: str, food_name: str) -> Union[int, NoReturn]:
    """
    Return how often a pet needs to be fed until it turns into a mount.

    :param pet_name: the pet to be fed
    :param food_name: the food to give the pet
    :return: times to feed, or exit if feeding this pet is not possible.
    """
    user: HabiticaUser = HabiticaUserRequest().request_user_data_or_exit()
    zoo: Zoo = ZooBuilder(user).build()
    pair: PetMountPair = zoo.get(pet_name)
    if pair is None or pair.pet_available() is False:
        sys.exit(f"Can't feed pet {pet_name}. You don't have this pet.")
    if pair.mount_available():
        sys.exit(f"Can't feed pet {pet_name}. You have the mount.")

    pet: Pet = pair.pet
    if pair.can_feed_pet() is False:
        sys.exit(f"Can't feed pet {pet_name}. "
                 + pet.feed_status_explanation())

    return pet.required_food_items_until_mount(food_name)


def get_appropriate_food_or_exit(pet_name: str) -> Union[str, NoReturn]:
    """Return the food that is appropriate for this pet.

    For pets hatched with normal hatching potion, return their favorite food.
    For pets hatched with magic hatching potion, return the most abundant food you have.
    Other types of pets should not be received by this function.
    """
    pet = Pet(pet_name)
    if pet.has_just_1_favorite_food():
        return pet.favorite_food()

    if pet.likes_all_food():
        user: HabiticaUser = HabiticaUserRequest().request_user_data_or_exit()
        stockpile: FoodStockpile = FoodStockpileBuilder().user(user).build()
        return stockpile.get_most_abundant_food()

    # This should be unreachable if we configured click with the correct pets.
    msg = (f"We tried to find the appropriate food for {pet_name=}.\n"
           "We assumed that we would have found the appropriate food by now.\n"
           "Unfortunately, that was not the case.")
    raise YouFoundABugRewardError(msg)


_TIMES_OPTION = "--times"
_UNTIL_MOUNT_OPTION = "--until-mount"


@click.command()
@click.argument(
    "pet_name", type=click.Choice(PetData.feedable_pet_names),
    metavar="PET_NAME"
)
@click.argument(
    "food_name", type=click.Choice(FoodData.drop_food_names),
    metavar="[FOOD_NAME]", required=False
)
@click.option(
    _TIMES_OPTION, type=valid_feed_amount_range,
    metavar="N_FOOD",
    help="Number of FOOD_NAME to feed to PET_NAME.\n "
         f"Values under {MIN_FEED_TIMES} are clamped to {MIN_FEED_TIMES}. "
         f"Values over {MAX_FEED_TIMES} are clamped to {MAX_FEED_TIMES}."
)
@click.option(
    f"{_UNTIL_MOUNT_OPTION}/--no-until-mount",
    default=False, show_default=True,
    help="Keep feeding this pet until it is a mount."
)
@click.option(
    "--list-favorite-food/--no-list-favorite-food",
    default=False, show_default=True,
    help="Print favorite food for PET_NAME and exit."
)
def feed(pet_name: str, food_name: str,
         times: int, until_mount: bool,
         list_favorite_food: bool):
    """Feed a single pet.

     \b
     PET_NAME   name of the pet (e.g. Wolf-Golden)
     FOOD_NAME  name of the food (e.g. Honey).

     \b
     Examples:
     ---
     # Feed a Beetle-Skeleton a Fish
     $ hopla feed Beetle-Skeleton Fish

     \b
     # Feed a pet its favorite food once. This only works if a pet
     # pet is feedable. For magic hatched pets, the most
     # abundant food you have is used.
     $ hopla feed Rock-White

     \b
     # Feed a pet its favorite food until it is a mount. This
     # only works if a pet is feedable.
     $ hopla feed TRex-Red --until-mount

     \b
     # Feed a pet a specific type of food until it is a mount. This
     # commands also work with pets that like all foods. (e.g. pets
     # hatched with magic hatching potions.)
     $ hopla feed --until-mount Wolf-SandSculpture RottenMeat

     \b
     # Feed a Snail-Desert 5 Potatoes. Note that this commands fails
     # to feed anything if fewer than 5 Potatoes are required
     # for a pet to become a mount. It also fails if you don't have
     # this pet, or not enough potatoes.
     $ hopla feed --times=5 Snail-Desert Potatoe

     \b
     # List a pet's favorite food
     $ hopla feed Axolotl-Base --list-favorite-food
     Meat

     \b
     # Tip: You can use the <Tab> key to show the pet and food keys
     $ hopla feed <Tab><Tab>
     $ hopla feed Axolotl-White <Tab><Tab>
     $ hopla feed Axolotl-White Milk

     [API-docs](https://habitica.com/apidoc/#api-User-UserFeed)
    \f
    :return:

    Note: this API endpoint expect 'amount' as a query params (?amount=N) instead
    of a request body (even though it is a HTTP POST).
    """
    log.debug(f"hopla feed {pet_name=}, {food_name=}"
              f" {times=} {until_mount=}"
              f" {list_favorite_food=}")

    FeedCommandParameterChecker(times=times, until_mount=until_mount) \
        .raise_if_conflicting_feed_time_options()

    if list_favorite_food:
        print_favorite_food_and_exit(pet_name=pet_name)

    if food_name is None:
        food_name = get_appropriate_food_or_exit(pet_name=pet_name)
        log.debug(f"Food is automatically selected to be {food_name=}.")

    if until_mount:
        times: int = get_feed_times_until_mount(pet_name=pet_name,
                                                food_name=food_name)
    else:
        times: int = times or 1

    pet_feed_request = FeedPostRequester(
        pet_name=pet_name,
        food_name=food_name,
        food_amount=times
    )

    response: requests.Response = pet_feed_request.post_feed_request()
    return get_feed_data_or_exit(feed_response=response)


@dataclass
class FeedCommandParameterChecker:
    """
    A class that raises errors when `hopla feed` received an invalid
    combination of parameters.
    """
    times: Optional[int]
    until_mount: bool

    def raise_if_conflicting_feed_time_options(self) -> Optional[NoReturn]:
        """
        The user should not use conflicting options that both specify
        how many food items should be given.
        """
        if self.until_mount and self.times is not None:
            raise click.UsageError(
                "Conflicting options: "
                f"Cannot specify both {_TIMES_OPTION} and {_UNTIL_MOUNT_OPTION}."
            )



"""
The module with CLI code that handles the `hopla set` group command.
"""
import logging

import click
import requests

from hopla.hoplalib.http import RequestHeaders, UrlBuilder
from hopla.hoplalib.outputformatter import JsonFormatter
from hopla.hoplalib.http import HabiticaRequest

log = logging.getLogger()


@click.group()
def set():  # pylint: disable=redefined-builtin
    """
    GROUP to set things in Habitica.
    """


@set.command()
@click.argument("day_start_hour", type=click.IntRange(min=0, max=23),
                default=0, metavar="[HOUR]")
@click.option("--json/--no-json", "json_flag",
              default=False, show_default=True,
              help="Output the command result in JSON.")
def day_start(day_start_hour, json_flag: bool):
    """Set your day-start (CRON) to the specified HOUR.

    HOUR - sets the day start to the N-th hour of the day. HOUR may range
    from 0 to 23 (inclusive). 0 AM (midnight) will be used as the
    default day start hour when no hour is specified.


    \b
    Examples:
    ---
    # Set the day start to 01:00 AM.
    $ hopla set day-start 1
    Your custom day start has changed.

    \b
    # Set the day start to 00:00 AM and get JSON output (e.g. for pipelines).
    $ hopla set day-start --json
    { "message": "Your custom day start has changed." }


    \f
    :param day_start_hour:
    :param json_flag
    """
    log.debug(f"hopla set day-start {day_start_hour}")

    headers = RequestHeaders().get_default_request_headers()
    body = {"dayStart": day_start_hour}
    url = UrlBuilder(path_extension="/user/custom-day-start").url

    response = requests.post(
        url=url,
        headers=headers,
        json=body,
        timeout=HabiticaRequest.TIMEOUT
    )

    json = response.json()
    json_data = json["data"]
    if json_flag:
        click.echo(JsonFormatter(json_data).format_with_double_quotes())
    else:
        click.echo(json_data["message"])



#!/usr/bin/env python3
import random
from typing import List

import pytest
from click import ClickException

from hopla.hoplalib.common import GlobalConstants
from hopla.hoplalib.errors import YouFoundABugRewardError
from hopla.hoplalib.zoo.fooddata import FoodData
from hopla.hoplalib.zoo.foodmodels import FeedStatus, InvalidFeedStatus
from hopla.hoplalib.zoo.petdata import PetData
from hopla.hoplalib.zoo.petmodels import InvalidPet, InvalidPetMountPair, Mount, Pet, PetMountPair

_SAMPLE_SIZE = 25
"""
Sample size to use for pytest.mark.parameterize when unittests become slow
"""


class TestPet:
    class TestPetInit:
        def test_init_raises_invalid_pet_name(self):
            with pytest.raises(InvalidPet) as execinfo:
                invalid_pet_name = "INVALID_PET"
                Pet(invalid_pet_name)

            err_msg = str(execinfo.value)
            assert invalid_pet_name in err_msg
            assert f"an issue at {GlobalConstants.NEW_ISSUE_URL}" in err_msg

        @pytest.mark.parametrize(
            "pet_name",
            # Testing all pets takes too long (approx. 2 seconds), so we sample.
            random.sample(PetData.pet_names, k=_SAMPLE_SIZE)
        )
        def test_init_valid_pet_name_ok(self, pet_name: str):
            pet = Pet(pet_name=pet_name)
            assert pet.name == pet_name

        @pytest.mark.parametrize("invalid_feed_status", [-10, -2, 1, 3, 51, 100])
        def test_init_raises_invalid_feed_status_fail(self, invalid_feed_status: int):
            with pytest.raises(InvalidFeedStatus) as execinfo:
                feed_status = FeedStatus(invalid_feed_status)
                Pet("PandaCub-IcySnow", feed_status=feed_status)

            err_msg = str(execinfo.value)
            assert f"feed_status={int(invalid_feed_status)}" in err_msg

        @pytest.mark.parametrize("valid_feed_status", [5, 10, 20, -1, 49])
        def test_init_sets_valid_feed_status_ok(self, valid_feed_status: int):
            feed_status = FeedStatus(valid_feed_status)
            pet = Pet("LionCub-Sunset", feed_status=feed_status)

            assert valid_feed_status == int(pet.feed_status)

    class TestFavoriteFood:
        @pytest.mark.parametrize(
            "pet_name,expected_favorite_food", [
                ("Wolf-Base", "Meat"),
                ("FlyingPig-CottonCandyPink", "CottonCandyPink"),
                ("PandaCub-CottonCandyBlue", "CottonCandyBlue"),
                ("TigerCub-Golden", "Honey"),
                ("Fox-Red", "Strawberry"),
                ("BearCub-White", "Milk"),
                ("Dragon-Desert", "Potatoe"),
                ("Robot-Shade", "Chocolate"),
                ("Dolphin-Skeleton", "Fish")
            ]
        )
        def test_favorite_food_for_pets_with_favorite_food_ok(self, pet_name: str,
                                                              expected_favorite_food: str):
            pet = Pet(pet_name)
            assert pet.favorite_food() == expected_favorite_food

        @pytest.mark.parametrize(
            "pet_name", random.sample(PetData.only_1favorite_food_pet_names, k=_SAMPLE_SIZE)
        )
        def test_favorite_food_single_favorite_food_ok(self, pet_name: str):
            pet = Pet(pet_name)
            alternative_food = "SomeFakeFood"

            assert "Any" != pet.favorite_food()
            assert alternative_food != pet.favorite_food(
                default_value_for_all_favorite_food=alternative_food
            )
            assert "Unfeedable" != pet.favorite_food(
                default_value_for_unfeedable=alternative_food
            )

        @pytest.mark.parametrize(
            # only magic potion pets like all foods
            "pet_name", random.sample(PetData.magic_potion_pet_names, k=_SAMPLE_SIZE)
        )
        def test_favorite_food_multiple_favorite_food_ok(self, pet_name: str):
            pet = Pet(pet_name)

            alternative_food = "Whatever"
            default_value_for_all_favorite_food_ok_pets = "Any"
            assert alternative_food == pet.favorite_food(
                default_value_for_all_favorite_food=alternative_food
            )
            assert default_value_for_all_favorite_food_ok_pets == pet.favorite_food(
                default_value_for_unfeedable=alternative_food
            )
            assert default_value_for_all_favorite_food_ok_pets == pet.favorite_food()

        @pytest.mark.parametrize(
            # only magic potion pets like all foods
            "pet_name", PetData.unfeedable_pet_names
        )
        def test_favorite_food_for_unfeedable_pets(self, pet_name: str):
            pet = Pet(pet_name)

            alternative_food = "Whatever"
            default_value_for_unfeedable_pets = "Unfeedable"
            assert default_value_for_unfeedable_pets == pet.favorite_food(
                default_value_for_all_favorite_food=alternative_food
            )
            assert alternative_food == pet.favorite_food(
                default_value_for_unfeedable=alternative_food)
            assert default_value_for_unfeedable_pets == pet.favorite_food()

    class TestIsFavoriteFood:

        @pytest.mark.parametrize(
            "pet_name,food_name",
            # we use choices here because both groups are smaller than
            # the specified k
            zip(random.choices(PetData.unfeedable_pet_names, k=_SAMPLE_SIZE),
                random.choices(FoodData.drop_food_names, k=_SAMPLE_SIZE))
        )
        def test_is_favorite_food_false_because_unfeedable(self, pet_name: str,
                                                           food_name: str):
            pet = Pet(pet_name)
            assert pet.is_favorite_food(food_name) is False

        @pytest.mark.parametrize(
            "pet_name,food_name",
            zip(random.sample(PetData.magic_potion_pet_names, k=_SAMPLE_SIZE),
                random.choices(FoodData.drop_food_names, k=_SAMPLE_SIZE))
        )
        def test_is_favorite_food_true_because_likes_all_food(self,
                                                              pet_name: str,
                                                              food_name: str):
            assert Pet(pet_name).is_favorite_food(food_name)

        @pytest.mark.parametrize(
            "pet_name,food_name,expected_result", [
                ("Fox-Red", "Strawberry", True),
                ("Fox-Red", "Milk", False),
                ("Parrot-Desert", "Potatoe", True),
                ("Parrot-Base", "Potatoe", False),
                ("Hippo-Golden", "Honey", True),
                ("Hippo-Golden", "CottonCandyBlue", False),
            ]
        )
        def test_is_favorite_food_for_hatch_pets(self,
                                                 pet_name: str,
                                                 food_name: str,
                                                 expected_result: bool):
            pet = Pet(pet_name)
            result = pet.is_favorite_food(food_name)
            assert result is expected_result

    class TestOtherPetFunctions:
        """ Test class for Pet functions that aren't covered elsewhere """
        PET_NAME = "Fox-Ember"

        def test___repr__(self):
            pet_name = "Fox-Shadow"
            feed_status = FeedStatus(10)
            pet = Pet(pet_name, feed_status=feed_status)

            result = str(pet)

            expected = f"Pet({pet_name}: {feed_status})"
            assert result == expected

        @pytest.mark.parametrize(
            "feed_status,partial_expected_message", [
                (-1, f"can't feed self.name='{PET_NAME}' you only have the mount"),
                (0, f"You released {PET_NAME}, so you cannot feed it"),
                (5, f"Cannot determine if self.name='{PET_NAME}' can be fed"),
                (20, f"name='{PET_NAME}' can be fed")
            ]
        )
        def test_feed_status_explanation_ok(self, feed_status: int,
                                            partial_expected_message: str):
            feed_status = FeedStatus(feed_status)
            pet = Pet(TestPet.TestOtherPetFunctions.PET_NAME, feed_status=feed_status)

            result_explanation: str = pet.feed_status_explanation()

            assert partial_expected_message in result_explanation

        @pytest.mark.parametrize(
            "unfeedable_pet_name", [
                "Fox-Veteran", "JackOLantern-Glow", "JackOLantern-RoyalPurple",
                "MantisShrimp-Base",
                "Gryphon-Gryphatrice", "PandaCub-Dessert", "LionCub-Veggie", "LionCub-Dessert",
                "Fox-Veggie", "Fox-Dessert",
            ]
        )
        def test_feed_status_unfeedable_pet(self, unfeedable_pet_name: str):
            pet = Pet(unfeedable_pet_name)

            result_explanation: str = pet.feed_status_explanation()

            assert f"{pet.name} is an unfeedable pet." == result_explanation

        @pytest.mark.parametrize(
            "pet_name,is_gen1_expected", [
                ("Wolf-Base", True),
                ("FlyingPig-CottonCandyPink", True),
                ("PandaCub-CottonCandyBlue", True),
                ("TigerCub-Golden", True),
                ("Fox-Red", True),
                ("BearCub-White", True),
                ("Dragon-Desert", True),
                ("Cactus-Amber", False),
                ("BearCub-Frost", False),
                ("Butterfly-Red", False)
            ]
        )
        def test_is_generation1_pet_ok(self, pet_name, is_gen1_expected):
            pet = Pet(pet_name)
            assert pet.is_generation1_pet() is is_gen1_expected

        @pytest.mark.parametrize(
            "pet_name",
            random.sample(PetData.generation1_pet_names + PetData.quest_pet_names,
                          k=_SAMPLE_SIZE)
        )
        def test_has_just_1_favorite_food_true(self, pet_name):
            pet = Pet(pet_name=pet_name)
            assert pet.has_just_1_favorite_food()

        @pytest.mark.parametrize(
            "pet_name",
            random.sample(PetData.magic_potion_pet_names, k=_SAMPLE_SIZE)
        )
        def test_has_just_1_favorite_food_false(self, pet_name):
            pet = Pet(pet_name=pet_name)
            assert pet.has_just_1_favorite_food() is False

        @pytest.mark.parametrize(
            "pet_name,likes_all_food_expected", [
                ("Fox-Golden", False),
                ("LionCub-Holly", True),
                ("PandaCub-Cupid", True),
                ("Octopus-Red", False),
                ("Seahorse-CottonCandyBlue", False)
            ]
        )
        def test_likes_all_food(self, pet_name: str,
                                likes_all_food_expected: bool):
            pet = Pet(pet_name)

            result = pet.likes_all_food()

            assert result is likes_all_food_expected

        @pytest.mark.parametrize(
            "pet_name,expected_from_drop_hatch_potion", [
                ("Fox-Golden", True),
                ("Dragon-Shade", True),
                ("LionCub-Holly", False),
                ("PandaCub-Cupid", False),
                ("Octopus-Red", True),
                ("Cheetah-CottonCandyPink", True),
                ("Seahorse-CottonCandyBlue", True),
                ("Unicorn-Skeleton", True),
                ("MagicalBee-Base", False),  # although -Base postfix, this is a world boss reward
                ("Phoenix-Base", False),  # although -Base postfix, this is a world boss reward
                ("Turkey-Base", False),  # although -Base postfix, this is a event sequence reward
                ("Gryphon-Gryphatrice", False),
            ]
        )
        def test_is_from_drop_hatch_potion(self, pet_name: str,
                                           expected_from_drop_hatch_potion: bool):
            pet = Pet(pet_name)

            result = pet.is_from_drop_hatch_potions()

            assert result is expected_from_drop_hatch_potion

        @pytest.mark.parametrize(
            "pet_name", random.sample(PetData.quest_pet_names, k=_SAMPLE_SIZE)
        )
        def test_is_quest_pet_true(self, pet_name: str):
            pet = Pet(pet_name)
            assert pet.is_quest_pet() is True

        @pytest.mark.parametrize(
            "pet_name",
            random.sample(PetData.generation1_pet_names + PetData.magic_potion_pet_names,
                          k=_SAMPLE_SIZE)
        )
        def test_is_quest_pet_false(self, pet_name: str):
            pet = Pet(pet_name)
            assert pet.is_quest_pet() is False

        @pytest.mark.parametrize(
            "pet_name",
            random.sample(PetData.magic_potion_pet_names, k=_SAMPLE_SIZE)
        )
        def test_is_magic_hatch_potion_true(self, pet_name: str):
            pet = Pet(pet_name)
            assert pet.is_magic_hatch_pet() is True

        @pytest.mark.parametrize(
            "pet_name",
            random.sample(PetData.generation1_pet_names + PetData.rare_pet_names +
                          PetData.quest_pet_names,
                          k=_SAMPLE_SIZE)
        )
        def test_is_magic_hatch_potion_false(self, pet_name: str):
            pet = Pet(pet_name)
            assert pet.is_magic_hatch_pet() is False

        @pytest.mark.parametrize(
            "pet,food_name,expected_times", [
                (Pet("Hedgehog-Shade", feed_status=FeedStatus(10)), "Chocolate", 8),
                (Pet("Hedgehog-Shade", feed_status=FeedStatus(10)), "Milk", 20),
                (Pet("Egg-Base"), "Meat", 9),
                (Pet("Egg-Zombie"), "CottonCandyPink", 23),
                (Pet("Egg-Base", feed_status=FeedStatus(40)), "Meat", 2),
                (Pet("Egg-Base", feed_status=FeedStatus(44)), "Meat", 2),
                (Pet("Egg-Golden", feed_status=FeedStatus(47)), "Fish", 2),
            ]
        )
        def test_required_food_items_until_mount(self, pet: Pet,
                                                 food_name: str,
                                                 expected_times: int):
            times_result: int = pet.required_food_items_until_mount(food_name)

            assert times_result == expected_times


class TestMount:

    @pytest.mark.parametrize(
        "mount,expected", [
            (Mount("Wolf-Veteran", availability_status=True), "Mount(Wolf-Veteran: True)"),
            (Mount("Monkey-Shade", availability_status=None), "Mount(Monkey-Shade: None)")
        ]
    )
    def test__repr__(self, mount: Mount, expected: str):
        result: str = repr(mount)
        assert result == expected

    @pytest.mark.parametrize("mount,expected_available", [
        (Mount("Aether-Invisible", availability_status=True), True),
        (Mount("BearCub-Celestial", availability_status=False), False),
        (Mount("Cactus-Ember", availability_status=None), False)
    ])
    def test_is_available_ok(self, mount: Mount, expected_available: bool):
        assert mount.is_available() is expected_available


class TestPetMountPair:
    empty_pair = PetMountPair(pet=None, mount=None)

    phoenix_pet_only_pair = PetMountPair(pet=Pet("Phoenix-Base", feed_status=FeedStatus(5)),
                                         mount=None)

    phoenix_mount_only_pair = PetMountPair(pet=None,
                                           mount=Mount("Phoenix-Base", availability_status=True))

    mount_available_pairs: List[PetMountPair] = [
        # only mount
        PetMountPair(pet=Pet("Owl-Shade", feed_status=FeedStatus(-1)),
                     mount=Mount("Owl-Shade", availability_status=True)),
        # pet is there, and mount as well
        PetMountPair(pet=Pet("Egg-White", feed_status=FeedStatus(5)),
                     mount=Mount("Egg-White", availability_status=True)),
        phoenix_mount_only_pair
    ]

    unfeedable_pairs: List[PetMountPair] = [
        # no pets
        empty_pair, phoenix_mount_only_pair,
        # cannot feed pet if there is a mount
        *mount_available_pairs,
        # released pet
        PetMountPair(pet=Pet("Wolf-Base", feed_status=FeedStatus(0)),
                     mount=None),
        # unfeedable pet
        phoenix_pet_only_pair,
        # no corresponding pet for this mount
        PetMountPair(pet=None, mount=Mount("Aether-Invisible", availability_status=True))
    ]

    feedable_pairs: List[PetMountPair] = [
        # only pet
        PetMountPair(pet=Pet("Rat-CottonCandyBlue", feed_status=FeedStatus(10)),
                     mount=None),
        # mount was released
        PetMountPair(pet=Pet("Rat-CottonCandyPink", feed_status=FeedStatus(5)),
                     mount=Mount("Rat-CottonCandyPink", availability_status=None)),
    ]

    @pytest.mark.parametrize("pair", feedable_pairs + unfeedable_pairs)
    def test_init__ok(self, pair: PetMountPair):
        pet: Pet = pair.pet
        mount: Mount = pair.mount

        if pet is not None:
            assert pet.name is not None
            assert pet.feed_status is not None

        if mount is not None:
            assert mount.name is not None
            assert mount.is_available() in [True, False, None]

    def test__init__pet_mount_diff_names_fail(self):
        pet_name = "Rat-CottonCandyBlue"
        mount_name = "Sloth-Red"
        with pytest.raises(InvalidPetMountPair) as exec_info:
            PetMountPair(pet=Pet(pet_name),
                         mount=Mount(mount_name, availability_status=None))

        err_mesg = str(exec_info.value)
        assert err_mesg.startswith(f"pet name '{pet_name}' must equal mount name '{mount_name}'")
        assert exec_info.errisinstance(ClickException)
        assert exec_info.errisinstance(YouFoundABugRewardError)
        assert exec_info.value.exit_code == 1

    @pytest.mark.parametrize("pair", unfeedable_pairs)
    def test_can_feed_pet_false(self, pair: PetMountPair):
        assert pair.can_feed_pet() is False

    @pytest.mark.parametrize("pair", feedable_pairs)
    def test_can_feed_pet_true(self, pair: PetMountPair):
        assert pair.can_feed_pet() is True

    # if pet is feedable, there cannot be a mount
    @pytest.mark.parametrize("pair", [empty_pair] + feedable_pairs)
    def test_mount_available_false(self, pair: PetMountPair):
        assert pair.mount_available() is False

    @pytest.mark.parametrize("pair", mount_available_pairs)
    def test_mount_available_true(self, pair: PetMountPair):
        assert pair.mount_available() is True

    @pytest.mark.parametrize("pair", [phoenix_pet_only_pair] + feedable_pairs)
    def test_pet_available_true(self, pair):
        assert pair.pet_available() is True

    @pytest.mark.parametrize("pair", [empty_pair, phoenix_mount_only_pair])
    def test_pet_available_false(self, pair):
        assert pair.pet_available() is False



#!/usr/bin/env python3
import pytest

from hopla.hoplalib.hatchery.hatchcontroller import HatchRequester


class TestHatchRequester:
    @pytest.mark.parametrize("egg_name,potion_name", [
        ("Cheetah", "Base"),
        ("Ferret", "CottonCandyBlue"),
        ("Wolf", "Skeleton"),
        ("Fox", "Polkadot")
    ])
    def test_url(self, egg_name: str, potion_name: str):
        requester = HatchRequester(egg_name=egg_name, hatch_potion_name=potion_name)

        result_url: str = requester.url

        assert result_url.endswith(f"/user/hatch/{egg_name}/{potion_name}")



#!/usr/bin/env python3
"""
Module with common click options and arguments.
"""
from typing import Final, List

import click

NO_INTERACTION_OPTION_NAMES: Final[List[str]] = ["--force", "--yes", "-f"]


def no_interactive_option() -> click.option:
    """A decorator to handle --force consistently throughout hopla."""
    return click.option(
        *NO_INTERACTION_OPTION_NAMES, "no_interactive",
        is_flag=True, default=False, show_default=True,
        help="Don't ask for confirmation before executing this command."
    )



"""
A helper module with Pet Data.

Later on this module may be refactored to dynamically retrieve pet
data from the API. The retrieved data from the API can then be
cached locally. We then need to refresh our cache only when the
appVersion of the Habitica API increases. However, this approach
is overkill for now.
"""
from typing import List

from hopla.hoplalib.hatchery.egg_data import EggData
from hopla.hoplalib.hatchery.hatchpotion_data import HatchPotionData


def _combine_pets_and_pots(pets: List[str], pots: List[str]):
    return [pet + "-" + pot for pet in pets for pot in pots]


class PetData:
    """ Helper class with valid Habitica pets.

    For more info:
    * hopla api content | jq .petInfo
    * hopla api content | jq .mountInfo

    And also interesting:
    * hopla api content | jq .premiumPets
    * hopla api content | jq .pets
    * hopla api content | jq .questPets
    * hopla api content | jq .specialPets
    * hopla api content | jq .wackyPets
    """

    generation1_pet_names = _combine_pets_and_pots(
        pets=EggData.drop_egg_names,
        pots=HatchPotionData.drop_hatch_potion_names
    )
    """ https://habitica.fandom.com/wiki/Pets#Generation_1_Pets """

    # The Magic potion pets are known as "premium" by the API
    # <https://habitica.fandom.com/wiki/Pets#Magic_Potion_Pets>
    # hopla api content | jq '[.petInfo[] | select(.type=="premium")]'
    # hopla api content | jq '[.petInfo[] | select(.type=="premium") | .key]'
    magic_potion_pet_names = _combine_pets_and_pots(
        pets=EggData.drop_egg_names,
        pots=HatchPotionData.magic_hatch_potion_names
    )
    """Magic hatching pet names"""

    # Quest pets: @see:
    # <https://habitica.fandom.com/wiki/Pets#Quest_Pets>
    # <https://habitica.fandom.com/wiki/Pets#Magic_Potion_Pets>
    # hopla api content | jq '[.petInfo[] | select(.type=="quest")]'
    quest_pet_names = _combine_pets_and_pots(
        pets=EggData.quest_egg_names,
        pots=HatchPotionData.drop_hatch_potion_names
    )

    # hopla api content | jq '[.petInfo[] | select(.type=="wacky")]'
    wacky_pet_names = _combine_pets_and_pots(
        pets=EggData.drop_egg_names,
        pots=HatchPotionData.wacky_hatch_potion_names
    )

    world_boss_reward_pet_names = [
        "Hippogriff-Hopeful", "MagicalBee-Base", "Phoenix-Base", "Mammoth-Base",
        "MantisShrimp-Base"
    ]

    event_sequence_pet_names = [
        "Wolf-Veteran", "Turkey-Base", "JackOLantern-Base", "Tiger-Veteran", "Turkey-Gilded",
        "Lion-Veteran", "Gryphon-RoyalPurple", "JackOLantern-Ghost", "Orca-Base", "Bear-Veteran",
        "Fox-Veteran", "JackOLantern-Glow", "JackOLantern-RoyalPurple"
    ]
    """https://habitica.fandom.com/wiki/Event_Item_Sequences"""

    other_pet_names = [
        "Jackalope-RoyalPurple", "BearCub-Polar", "Dragon-Hydra", "Wolf-Cerberus",
        "Gryphatrice-Jubilant", "Gryphon-Gryphatrice", "Aether-Invisible"
    ]

    rare_pet_names = world_boss_reward_pet_names + event_sequence_pet_names + other_pet_names
    """Rare pet names are pets that cannot be hatched.
    @see: <https://habitica.fandom.com/wiki/Pets#Rare_Pets>
    """

    only_1favorite_food_pet_names = generation1_pet_names + quest_pet_names
    """Only quest pets and gen1 pets have only 1 favorite food (anno Sept. 2021)"""

    feedable_pet_names = only_1favorite_food_pet_names + magic_potion_pet_names
    """
    Pets that can be fed are the gen1 pets, quest pets, and magic potion pets. (anno Sept. 2021).
    """

    unfeedable_pet_names = rare_pet_names + wacky_pet_names
    """Only wacky pets and rare pets cannot be fed anno 2022."""

    pet_names = (generation1_pet_names
                 + magic_potion_pet_names
                 + quest_pet_names
                 + wacky_pet_names
                 + rare_pet_names)
    """ @see: hopla api content | jq '[.petInfo[] | .key]' """



#!/usr/bin/env python3
"""
Module with models for a Habitica user.
"""
from dataclasses import dataclass
from typing import Dict, Optional


@dataclass(frozen=True)
class HabiticaUser:
    """
    Class representing a habitica user.
    """
    # pylint: disable=too-many-public-methods
    #################################################################
    # In a way, pylint is right. This class does have a lot of functions.
    # Most of these are oneliners, however we can also create
    # UserInventory, UserStats, etc object to compose this class from.
    # We can also use an extension (e.g. HoplaUser) which isA HabiticaUser
    # but where we add 'extra' functionality such as get_gems.
    # However, that is really overkill for now, and really urgent.
    # Therefore, too-many-public-methods is disabled for this class.
    #################################################################

    user_dict: dict
    """
    The user_dict is assumed to be returned from a 200 ok requests.Response
    (and using Response.json()) when calling the /user endpoint and getting
    "data" from the json.
    """

    def __getitem__(self, key):
        return self.user_dict.__getitem__(key)

    def get_stats(self) -> dict:
        """Index the user_dict for 'stats' and return the result"""
        return self["stats"]

    def get_gp(self) -> float:
        """Get the gold of a user."""
        return self.get_stats()["gp"]

    def get_mp(self) -> float:
        """Get the mana of a user."""
        return self.get_stats()["mp"]

    def get_hp(self) -> float:
        """Get the health of a user."""
        return self.get_stats()["hp"]

    def get_inventory(self) -> dict:
        """Index the user_dict for 'items' and return the result"""
        return self["items"]

    def get_pets(self) -> Dict[str, int]:
        """Return the pets of a user.

        :return: A dictionary with pet_name as key and feed_status as value.
        For example: {"Spider-Base": -1, "TRex-Skeleton": 5}
        """
        return self.get_inventory()["pets"]

    def get_eggs(self) -> Dict[str, int]:
        """Return the eggs of a user.

        :return: A dictionary with egg_names as keys and amount as values.
        For example: { "Dragon": 338, "Octopus": 0}
        """
        return self.get_inventory()["eggs"]

    def get_hatch_potions(self) -> Dict[str, int]:
        """Return the hatching potions of a user.

        :return: A dict with hatch_potion_names as keys and amount as value.
        For example: { "Desert": 456, "Glow": 0}
        """
        return self.get_inventory()["hatchingPotions"]

    def get_mounts(self) -> Dict[str, Optional[bool]]:
        """Return the mounts of a user.

        :return: A dictionary with mount_name as key and availability as value.
        For example: { "Dragon-Base": None, "Octopus-Shade": True }
        """
        return self.get_inventory()["mounts"]

    def get_food(self) -> Dict[str, int]:
        """Return the food that the user has"""
        return self.get_inventory()["food"]

    def get_auth(self) -> dict:
        """Index the user_dict for 'auth' and return the result"""
        return self["auth"]

    def get_gems(self) -> int:
        """Get the number of gems of a user.

        gems are stored in the 'balance' field. 1 'balance' equals 4 gems
        [see](https://habitica.fandom.com/wiki/Gems#Information_for_Developers)
        """
        balance = self["balance"]
        return int(balance * 4)



#!/usr/bin/env python3
import random
from typing import List

import click
import pytest

from hopla.hoplalib.errors import YouFoundABugRewardError
from hopla.hoplalib.hatchery.eggmodels import Egg, EggCollection, EggException
from hopla.hoplalib.hatchery.egg_data import EggData
from hopla.hoplalib.hatchery.hatchpotion_data import HatchPotionData
from hopla.hoplalib.hatchery.hatchpotionmodels import HatchPotion

_SAMPLE_SIZE = 9


class TestEgg:
    def test__init__invalid_name_fail(self):
        name = "InvalidName"
        with pytest.raises(EggException) as exec_info:
            Egg(name)

        assert str(exec_info.value).startswith(f"{name} is not a valid egg name.")
        assert exec_info.errisinstance((YouFoundABugRewardError, click.ClickException))

    @pytest.mark.parametrize(
        "egg_name,quantity",
        list(zip(random.sample(EggData.egg_names, k=_SAMPLE_SIZE),
                 range(-_SAMPLE_SIZE, 0)))
    )
    def test__init__invalid_quantity_fail(self, egg_name: str, quantity: int):
        with pytest.raises(EggException) as exec_info:
            Egg(egg_name, quantity=quantity)

        assert str(exec_info.value).startswith(f"{quantity} is below 0.")
        assert exec_info.errisinstance((YouFoundABugRewardError, click.ClickException))

    @pytest.mark.parametrize("egg1,egg2,expected_equals", [
        (Egg("Horse"), Egg("Horse"), True),
        (Egg("Seahorse", quantity=1), Egg("Seahorse"), True),
        (Egg("Snake", quantity=1), Egg("Snake", quantity=2), False),
        (Egg("Whale"), Egg("Rat"), False),
        (Egg("Rooster", quantity=0), Egg("TRex", quantity=0), False),
        (Egg("Penguin", quantity=1), Egg("Owl", quantity=4), False),
    ])
    def test__eq__(self, egg1: Egg, egg2: Egg, expected_equals: bool):
        assert (egg1 == egg2) is expected_equals

    def test__repr__(self):
        name, quantity = "Dolphin", 9
        egg = Egg(name, quantity=quantity)

        result: str = repr(egg)

        assert result == f"Egg({name}: {quantity})"

    @pytest.mark.parametrize("standard_egg_name", ["Wolf", "Fox", "FlyingPig", "Dragon"])
    def test_is_standard_egg(self, standard_egg_name: str):
        egg = Egg(standard_egg_name)

        assert egg.is_standard_egg() is True
        assert egg.is_quest_egg() is False

    @pytest.mark.parametrize("quest_egg_name", [
        "Dolphin", "Gryphon", "Deer", "Egg", "Seahorse", "Parrot"
    ])
    def test_is_quest_egg(self, quest_egg_name: str):
        egg = Egg(quest_egg_name)

        assert egg.is_standard_egg() is False
        assert egg.is_quest_egg() is True

    @pytest.mark.parametrize(
        "egg_name,potion_name",
        list(zip(EggData.drop_egg_names,
                 HatchPotionData.hatch_potion_names))
    )
    def test_can_be_hatched_by_standard_egg_true(self, egg_name: str,
                                                 potion_name: str):
        egg = Egg(egg_name)
        potion = HatchPotion(potion_name)

        # standard eggs can be hatched by any potion
        assert egg.can_be_hatched_by(potion) is True

    @pytest.mark.parametrize(
        "egg_name,potion_name",
        list(zip(EggData.quest_egg_names,
                 HatchPotionData.drop_hatch_potion_names))
    )
    def test_can_be_hatched_by_magic_egg_drop_potion_true(self,
                                                          egg_name: str,
                                                          potion_name: str):
        egg = Egg(egg_name)
        potion = HatchPotion(potion_name)

        # quest eggs can be hatched by standard potion
        assert egg.can_be_hatched_by(potion) is True

    @pytest.mark.parametrize(
        "egg_name,potion_name",
        list(zip(
            random.sample(EggData.quest_egg_names, k=_SAMPLE_SIZE),
            random.sample(HatchPotionData.non_drop_hatch_potion_names, k=_SAMPLE_SIZE)
        ))
    )
    def test_can_be_hatched_by_magic_egg_non_drop_potion_false(self,
                                                               egg_name: str,
                                                               potion_name: str):
        egg = Egg(egg_name)
        potion = HatchPotion(potion_name)

        # quest eggs cannot be hatched by non standard potion
        assert egg.can_be_hatched_by(potion) is False


class TestEggCollection:
    def test__init__empty__ok(self):
        collection = EggCollection()
        assert collection == EggCollection({})
        assert len(collection) == 0

    def test__init__ok(self):
        collection = EggCollection({"Wolf": 0, "Whale": 42, "Horse": 2})

        assert len(collection) == 3
        assert collection["Wolf"] == Egg("Wolf", quantity=0)
        assert collection["Whale"] == Egg("Whale", quantity=42)
        assert collection["Horse"] == Egg("Horse", quantity=2)

    def test__eq__(self):
        assert EggCollection() == EggCollection()
        assert EggCollection() == EggCollection({})
        assert EggCollection({"Cow": 1}) == EggCollection({"Cow": 1})
        assert EggCollection({"Cow": 1}) != EggCollection({"Cow": 2})
        assert EggCollection({"Cow": 1}) != EggCollection({"Rat": 1})
        assert EggCollection() != EggCollection({"Ferret": 1})

    def test_values_ok(self):
        collection = EggCollection({"Wolf": 1, "Whale": 41, "Horse": 2})

        generator = collection.values()
        assert next(generator) == Egg("Wolf", quantity=1)
        assert next(generator) == Egg("Whale", quantity=41)
        assert next(generator) == Egg("Horse", quantity=2)
        with pytest.raises(StopIteration):
            _ = next(generator)

    def test_values_as_list_ok(self):
        collection = EggCollection({"Wolf": 1, "Whale": 41, "Horse": 2})

        result: List[Egg] = list(collection.values())

        expected: List[Egg] = [
            Egg("Wolf", quantity=1), Egg("Whale", quantity=41), Egg("Horse", quantity=2)
        ]
        assert result == expected

    def test__iter__ok(self):
        collection = EggCollection({"Wolf": 1, "Whale": 42, "Horse": 2})

        iterator = iter(collection)

        assert next(iterator) == "Wolf"
        assert next(iterator) == "Whale"
        assert next(iterator) == "Horse"
        with pytest.raises(StopIteration):
            next(iterator)

    def test_remove_egg_not_available_faile(self):
        collection = EggCollection({"Whale": 1})

        not_found_egg = "Unicorn"
        with pytest.raises(EggException) as exec_info:
            collection.remove_egg(Egg(not_found_egg))

        expected_msg = f"{not_found_egg} was not in the collection "
        assert str(exec_info.value).startswith(expected_msg)

    def test_remove_egg_ok(self):
        egg1_name, egg1_quantity = "Wolf", 1
        egg2_name, egg2_quantity = "Whale", 42
        egg3_name, egg3_quantity = "Horse", 2
        collection = EggCollection(
            {egg1_name: egg1_quantity, egg2_name: egg2_quantity, egg3_name: egg3_quantity}
        )

        collection.remove_egg(Egg(egg1_name))
        collection.remove_egg(Egg(egg2_name))
        collection.remove_egg(Egg(egg3_name))

        assert collection[egg1_name] == Egg(egg1_name, quantity=egg1_quantity - 1)
        assert collection[egg2_name] == Egg(egg2_name, quantity=egg2_quantity - 1)
        assert collection[egg3_name] == Egg(egg3_name, quantity=egg3_quantity - 1)

    def test_get_standard_egg_collection_ok(self):
        egg1_name, egg1_quantity = "Wolf", 1
        egg2_name, egg2_quantity = "Whale", 42  # quest egg
        egg3_name, egg3_quantity = "Fox", 0
        egg4_name, egg4_quantity = "PandaCub", 69
        collection = EggCollection({
            egg1_name: egg1_quantity,
            egg2_name: egg2_quantity,
            egg3_name: egg3_quantity,
            egg4_name: egg4_quantity
        })

        result: EggCollection = collection.get_standard_egg_collection()

        expected = EggCollection({
            egg1_name: egg1_quantity, egg3_name: egg3_quantity, egg4_name: egg4_quantity
        })
        assert result == expected

    def test_get_quest_egg_collection_ok(self):
        egg1_name, egg1_quantity = "Wolf", 1  # not quest egg
        egg2_name, egg2_quantity = "Whale", 42
        egg3_name, egg3_quantity = "Horse", 0
        egg4_name, egg4_quantity = "Axolotl", 69
        collection = EggCollection({
            egg1_name: egg1_quantity,
            egg2_name: egg2_quantity,
            egg3_name: egg3_quantity,
            egg4_name: egg4_quantity
        })

        result: EggCollection = collection.get_quest_egg_collection()

        expected = EggCollection({
            egg2_name: egg2_quantity, egg3_name: egg3_quantity, egg4_name: egg4_quantity
        })
        assert result == expected



#!/usr/bin/env python3
from hopla.hoplalib.errors import PrintableException, YouFoundABugRewardError


class TestYouFoundABugRewardError:
    def test___init__(self):
        errmsg = "This request is too hot to handle"
        ex = YouFoundABugRewardError(errmsg)

        assert errmsg in str(ex)
        assert "report it over here:"
        assert "https://github.com/melvio/hopla/issues/new" in str(ex)


class TestPrintableException:

    def test___repr___and___str__(self):
        msg = "This request is too cold to handle!"

        ex = PrintableException(msg=msg)

        assert str(ex) == msg
        assert repr(ex) == f"PrintableException({msg})"

    def test___repr__and___str___child_class(self):
        class DNASequencingError(PrintableException):
            """Exception for our test"""

        err_msg = "Sequencing was above acceptable price of 2.50$"
        err = DNASequencingError(msg=err_msg)

        assert str(err) == err_msg
        assert repr(err) == f"DNASequencingError({err_msg})"



#!/usr/bin/env python3
"""
Module that constructs the CLI, handles configuration files initialization,
and sets up logging.
"""
import logging
import sys

import click

from hopla.cli.add.todo import todo
from hopla.cli.authenticate import authenticate
from hopla.cli.buy.enchanted_armoire import enchanted_armoire
from hopla.cli.cast import cast
from hopla.cli.complete import complete
from hopla.cli.config import config
from hopla.cli.feed import feed
from hopla.cli.feed_all import feed_all
from hopla.cli.get_group import get_group
from hopla.cli.get_user.auth import auth
from hopla.cli.get_user.info import info
from hopla.cli.get_user.inventory import inventory
from hopla.cli.get_user.stats import stats
from hopla.cli.groupcmds.add import add
from hopla.cli.groupcmds.api import api
from hopla.cli.groupcmds.buy import buy
from hopla.cli.groupcmds.get_user import get_user
from hopla.cli.groupcmds.hatch import hatch
from hopla.cli.groupcmds.set import set  # pylint: disable=redefined-builtin
from hopla.cli.hatch.quest_egg import quest_egg
from hopla.cli.hatch.standard_egg import standard_egg
from hopla.cli.hatch_all import hatch_all
from hopla.cli.request import request
from hopla.cli.support_development import support_development
from hopla.cli.version import version
from hopla.hoplalib.common import GlobalConstants
from hopla.hoplalib.configuration import ConfigInitializer, ConfigurationFileParser
from hopla.hoplalib.hoplaversion import HoplaVersion


def setup_logging() -> logging.Logger:
    """Setup python logging for the entire hopla project"""

    parsed_loglevel: str = ConfigurationFileParser().get_full_config_name(
        "cmd_all.loglevel",
        fallback="info"
    )

    loglevel_mapping = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR
    }

    # https://docs.python.org/3.8/howto/logging.html#logging-basic-tutorial
    logging.basicConfig(
        format="[%(levelname)s][%(filename)s|%(asctime)s] %(message)s",
        level=loglevel_mapping[parsed_loglevel],
        datefmt="%Y-%m-%dT%H:%M:%S"
    )

    return logging.getLogger(__package__)


log = setup_logging()
log.debug(f"start application with arguments: {sys.argv}")

HOPLA_CONTEXT_SETTINGS = {
    "help_option_names": ["-h", "--help"],  # add -h
    "auto_envvar_prefix": GlobalConstants.APPLICATION_NAME
}


@click.group(context_settings=HOPLA_CONTEXT_SETTINGS)
@click.version_option(version=HoplaVersion().semantic_version())
def hopla():
    """hopla - a command line interface (CLI) to interact with habitica.com"""


def organize_cli() -> None:
    """Attach the subgroups and subcommands to the top hopla group command"""
    # pylint: disable=too-many-statements

    # authenticate
    hopla.add_command(authenticate)

    # add
    hopla.add_command(add)
    add.add_command(todo)

    # api
    hopla.add_command(api)

    # buy
    hopla.add_command(buy)
    buy.add_command(enchanted_armoire)

    # cast
    hopla.add_command(cast)

    # complete
    hopla.add_command(complete)

    # config
    hopla.add_command(config)

    # feed
    hopla.add_command(feed)

    # feed-all
    hopla.add_command(feed_all)

    # get-group
    hopla.add_command(get_group)

    # get-user
    hopla.add_command(get_user)
    get_user.add_command(inventory)
    get_user.add_command(stats)
    get_user.add_command(info)
    get_user.add_command(auth)

    # hatch
    hopla.add_command(hatch)
    hatch.add_command(standard_egg)
    hatch.add_command(quest_egg)

    # hatch-all
    hopla.add_command(hatch_all)

    # request
    hopla.add_command(request)

    # support-development
    hopla.add_command(support_development)

    # version
    hopla.add_command(version)

    # set
    hopla.add_command(set)


def init_hopla_config_files() -> None:
    """Setup the config file."""
    had_to_create_new_config_file: bool = ConfigInitializer().initialize_before_running_cmds()
    if had_to_create_new_config_file is True:
        click.echo(f"Thank you for trying out {GlobalConstants.APPLICATION_NAME}")
        click.echo(
            "Bug reports, pull requests, and feature requests are welcomed over at:  "
        )
        click.echo(GlobalConstants.ISSUE_URL)


def kickstart_hopla() -> None:
    """Setup the config files, organize the CLI, and call the base command group."""
    init_hopla_config_files()
    organize_cli()
    hopla()



#!/usr/bin/env python3
"""
Module with hatching potion data.
"""
from typing import List


class HatchPotionData:
    """Class with data about hatching potions."""
    drop_hatch_potion_names = [
        "Base", "CottonCandyBlue", "CottonCandyPink", "Desert", "Golden",
        "Red", "Shade", "Skeleton", "White", "Zombie"
    ]
    """
    Drop hatching potions. Also known as 1st gen potions.
    This list was retrieved by using:
        hopla api content | jq .dropHatchingPotions | jq keys
    """

    magic_hatch_potion_names = [
        "Amber", "Aquatic", "Aurora", "AutumnLeaf",
        "BirchBark", "BlackPearl", "Bronze",
        "Celestial", "Cupid", "Ember",
        "Fairy", "Floral", "Fluorite", "Frost",
        "Ghost", "Glass", "Glow", "Holly",
        "IcySnow", "Moonglow", "MossyStone", "Onyx",
        "Peppermint", "PinkMarble", "PolkaDot", "Porcelain",
        "Rainbow", "RoseGold", "RoseQuartz", "RoyalPurple", "Ruby",
        "SandSculpture", "Shadow", "Shimmer", "Silver", "SolarSystem",
        "Spooky", "StainedGlass", "StarryNight", "Sunset", "Sunshine",
        "Thunderstorm", "Turquoise", "Vampire", "Watery", "Windup"
    ]
    """
    Magic hatching potions. This list was retrieved by using:
        hopla api content | jq .premiumHatchingPotions | jq keys
    """

    wacky_hatch_potion_names = ["Dessert", "TeaShop", "Veggie", "VirtualPet"]
    """
    Wacky hatching potions. This list was retrieved by using:
        hopla api content | jq '.wackyHatchingPotions|keys'
    """

    non_drop_hatch_potion_names: List[str] = (
            magic_hatch_potion_names
            + wacky_hatch_potion_names
    )

    hatch_potion_names: List[str] = drop_hatch_potion_names + non_drop_hatch_potion_names
    """All potion names in habitica."""



#!/usr/bin/env python3
import random

import pytest

from hopla.hoplalib.zoo.petdata import PetData

_SAMPLE_SIZE_LARGE = 20
_SAMPLE_SIZE_SMALL = 5


class TestPetData:

    @pytest.mark.parametrize(
        "gen1pet,magic_pet,quest_pet", list(zip(
            random.sample(PetData.generation1_pet_names, k=_SAMPLE_SIZE_LARGE),
            random.sample(PetData.magic_potion_pet_names, k=_SAMPLE_SIZE_LARGE),
            random.sample(PetData.quest_pet_names, k=_SAMPLE_SIZE_LARGE)
        ))
    )
    def test_correct_grouping_of_feedable_pets(self, gen1pet: str,
                                               magic_pet: str,
                                               quest_pet: str):
        assert gen1pet not in PetData.magic_potion_pet_names
        assert gen1pet not in PetData.quest_pet_names
        assert magic_pet not in PetData.generation1_pet_names
        assert magic_pet not in PetData.quest_pet_names
        assert quest_pet not in PetData.generation1_pet_names
        assert quest_pet not in PetData.magic_potion_pet_names

        assert gen1pet in PetData.feedable_pet_names
        assert magic_pet in PetData.feedable_pet_names
        assert quest_pet in PetData.feedable_pet_names

        assert gen1pet in PetData.only_1favorite_food_pet_names
        assert magic_pet not in PetData.only_1favorite_food_pet_names
        assert quest_pet in PetData.only_1favorite_food_pet_names

        assert gen1pet not in PetData.unfeedable_pet_names
        assert magic_pet not in PetData.unfeedable_pet_names
        assert quest_pet not in PetData.unfeedable_pet_names

        assert gen1pet in PetData.pet_names
        assert magic_pet in PetData.pet_names
        assert quest_pet in PetData.pet_names

    @pytest.mark.parametrize(
        "wacky_pet,world_boss_pet,event_sequence_pet,other_pet", list(zip(
            random.sample(PetData.wacky_pet_names, k=_SAMPLE_SIZE_SMALL),
            random.sample(PetData.world_boss_reward_pet_names, k=_SAMPLE_SIZE_SMALL),
            random.sample(PetData.event_sequence_pet_names, k=_SAMPLE_SIZE_SMALL),
            random.sample(PetData.other_pet_names, k=_SAMPLE_SIZE_SMALL),
        ))
    )
    def test_correct_grouping_rare_pets(self,
                                        wacky_pet: str,
                                        world_boss_pet: str,
                                        event_sequence_pet: str,
                                        other_pet: str):
        assert wacky_pet not in PetData.world_boss_reward_pet_names
        assert wacky_pet not in PetData.event_sequence_pet_names
        assert wacky_pet not in PetData.other_pet_names

        assert world_boss_pet not in PetData.wacky_pet_names
        assert world_boss_pet not in PetData.event_sequence_pet_names
        assert world_boss_pet not in PetData.other_pet_names

        assert event_sequence_pet not in PetData.wacky_pet_names
        assert event_sequence_pet not in PetData.world_boss_reward_pet_names
        assert event_sequence_pet not in PetData.other_pet_names

        assert other_pet not in PetData.wacky_pet_names
        assert other_pet not in PetData.event_sequence_pet_names
        assert other_pet not in PetData.world_boss_reward_pet_names

        assert wacky_pet in PetData.unfeedable_pet_names
        assert world_boss_pet in PetData.unfeedable_pet_names
        assert event_sequence_pet in PetData.unfeedable_pet_names
        assert other_pet in PetData.unfeedable_pet_names

        assert wacky_pet not in PetData.only_1favorite_food_pet_names
        assert world_boss_pet not in PetData.only_1favorite_food_pet_names
        assert event_sequence_pet not in PetData.only_1favorite_food_pet_names
        assert other_pet not in PetData.only_1favorite_food_pet_names

        assert wacky_pet not in PetData.feedable_pet_names
        assert world_boss_pet not in PetData.feedable_pet_names
        assert event_sequence_pet not in PetData.feedable_pet_names
        assert other_pet not in PetData.feedable_pet_names

        assert wacky_pet not in PetData.rare_pet_names
        assert world_boss_pet in PetData.rare_pet_names
        assert event_sequence_pet in PetData.rare_pet_names
        assert other_pet in PetData.rare_pet_names



#!/usr/bin/env python3
"""
Module with the command for the `hopla add todo` command.
"""
import logging
import sys
from datetime import datetime, timedelta
from typing import Final, List

import click

from hopla.hoplalib.outputformatter import JsonFormatter
from hopla.hoplalib.tasks.taskcontroller import AddTodoRequest
from hopla.hoplalib.tasks.taskmodel import HabiticaChecklist, HabiticaTodo, TaskDifficultyData

log = logging.getLogger()


def create_habitica_todo(*,
                         checklist_editor: bool,
                         checklist_file,
                         difficulty: str,
                         due_date: datetime,
                         todo_name: str) -> HabiticaTodo:
    """Create a HabiticaTodo object from the provided parameters"""
    habitica_checklist: HabiticaChecklist = get_checklist(
        checklist_file=checklist_file, checklist_editor=checklist_editor
    )
    if todo_name is None:
        todo_name: str = click.prompt("Please provide a name for your todo")
    return HabiticaTodo(
        todo_name=todo_name,
        difficulty=difficulty,
        due_date=due_date,
        checklist=habitica_checklist
    )


def get_checklist(checklist_file, checklist_editor: bool) -> HabiticaChecklist:
    """Get a habitica checklist from the user

     Returns an empty HabiticaChecklist if user did not want a checklist

     >>> get_checklist(None, False)
     HabiticaChecklist(checklist=[])
    """
    if checklist_editor:
        return get_checklist_with_editor(checklist_file)
    if checklist_file:
        return HabiticaChecklist(checklist=checklist_file.readlines())
    return HabiticaChecklist(checklist=[])


def get_checklist_with_editor(checklist_file) -> HabiticaChecklist:
    """
    Get the HabiticaUser for an user that requested the editor.
    """
    comment = "# Every line below this comment represents an item on your checklist\n"
    if checklist_file:
        text: str = comment + "".join(checklist_file.readlines())
    else:
        text: str = comment
    message = click.edit(text=text, extension=".md")
    if message is not None:
        try:
            checklist_str_with_comment: List[str] = message.split(comment, maxsplit=1)
        except IndexError as ex:
            sys.exit(f"could not find the checklist below the specified comment: {ex}")
        checklist_str = checklist_str_with_comment[1]
        return HabiticaChecklist(checklist=checklist_str.split("\n"))

    sys.exit("editor exited")


_DUE_DATE_OPTION_NAME: Final[str] = "--due-date"
_DUE_DATE_TODAY_FLAG: Final[str] = "--today"
_DUE_DATE_TOMORROW_FLAG: Final[str] = "--tomorrow"
__DUE_DATE_FEATURE_FLAG_PARAM: Final[str] = "due_date"

date_formats = ["%Y-%m-%d", "%d-%m-%Y"]
TODAY = datetime.today()
TOMORROW = TODAY + timedelta(days=1)


@click.command()
@click.option(
    "--difficulty",
    type=click.Choice(TaskDifficultyData.VALID_DIFFICULTIES),
    default="easy",
    show_default=True,
    help="the priority of the To-Do"
)
@click.option(
    _DUE_DATE_OPTION_NAME, "--deadline", __DUE_DATE_FEATURE_FLAG_PARAM,
    type=click.DateTime(formats=date_formats),
    help="Due date of the To-Do in format YYYY-MM-DD or DD-MM-YYYY."
)
@click.option(
    _DUE_DATE_TODAY_FLAG, __DUE_DATE_FEATURE_FLAG_PARAM, flag_value=TODAY,
    type=click.DateTime(formats=date_formats),
    help="Use today as the due date."
)
@click.option(
    _DUE_DATE_TOMORROW_FLAG, __DUE_DATE_FEATURE_FLAG_PARAM, flag_value=TOMORROW,
    type=click.DateTime(formats=date_formats),
    help="Use tomorrow as the due date.")
@click.option(
    "--checklist", "checklist_file",
    type=click.File(),
    help="Every line in FILENAME will be an item of the checklist."
)
@click.option(
    "--editor", "checklist_editor", is_flag=True,
    default=False,
    help="Open up an editor to create a checklist interactively",
    show_default=True
)
@click.argument("todo_name", required=False)
def todo(difficulty: str,
         due_date: datetime,
         checklist_file,
         checklist_editor: bool,
         todo_name: str):
    """Add a To-Do.

    TODO_NAME the name of the To-Do. When omitted, hopla will prompt the user to input
    a TODO_NAME interactively.

    \b
    Examples:
    ---
    # Simplest way to create a To-Do:
    $ hopla add todo "My todo"

    \b
    # If you don't specify a To-Do name, hopla will prompt you for one.
    $ hopla add todo
    > Please provide a name for your todo: My todo

    \b
    # Override the default 'easy' difficulty using either hard|medium|trivial.
    $ hopla add --difficulty=hard todo 'This is a hard todo'

    \b
    # --due-date and --deadline work as synonyms.
    $ hopla add todo --deadline 2042-01-21 "My Todo"
    $ hopla add todo --due-date 2042-01-21 "My Todo"

    \b
    # --due-date and --deadline support YYYY-MM-DD and DD-MM-YYYY, so the
    # following commands are the same:
    $ hopla add todo --deadline 12-31-2101 "My Todo"
    $ hopla add todo --deadline 2101-12-31 "My Todo"


    \b
    # Use the 'today' or 'tomorrow' to set the due date to today/tomorrow.
    $ hopla add todo --today "Feed pet"
    $ hopla add todo --tomorrow "Feed pet Again"

    \b
    # You can use cool shell tricks to provide checklists on the fly.
    # e.g. sort a list in a file:
    $ hopla add todo --checklist <(sort ./tasks.txt) "Sorted Checklist"

    \b
    # Open up an editor and type in your new checklist on the go.
    $ hopla add todo --editor "My Long Checklist"
    > editor opens

    \b
    # Open up a file as a template for your checklist, edit it, and send
    # it as a checklist.
    # The underlying file remains unchanged.
    $ hopla add todo --checklist tasks.txt --editor "My editable checklist"
    > editor opens tasks.txt


    \b
    # Create a hard To-Do with due date on the 22nd of october in 2077, and
    # use every line in the `my_todo.md` FILE as an item of your checklist.
    $ hopla add todo --difficulty hard --due-date=2077-10-21 \\
                   --checklist-file ./my_todo.md 'Text for my todo'


    [apidocs](https://habitica.com/apidoc/#api-Task-CreateUserTasks)

    \f
    :param checklist_editor:
    :param difficulty: str
    :param due_date:
    :param checklist_file:
    :param todo_name:
    :return:

    """
    log.debug(f"habitica add todo name={todo_name}\n"
              f"   difficulty={difficulty}, "
              f"   due_date={due_date},"
              f"   checklist ={checklist_file}, editor={checklist_editor}")

    habitica_todo: HabiticaTodo = create_habitica_todo(
        checklist_editor=checklist_editor,
        checklist_file=checklist_file,
        difficulty=difficulty,
        due_date=due_date,
        todo_name=todo_name
    )

    add_todo_response_data = AddTodoRequest(habitica_todo=habitica_todo).post_add_todo_request()

    click.echo(JsonFormatter(add_todo_response_data).format_with_double_quotes())



#!/usr/bin/env python3
import re
from re import Pattern
from typing import Optional

import pytest

from hopla.hoplalib.hoplaversion import HoplaVersion


class TestHoplaVersion:

    def test_semantic_version_defaults(self, semver_pattern: Pattern):
        hopla_version = HoplaVersion()
        result: str = hopla_version.semantic_version()
        assert re.fullmatch(pattern=semver_pattern, string=result)

    @pytest.mark.parametrize("prerelease_value", [None, ""])
    def test_semantic_version_without_prerelease(self, semver_pattern: Pattern,
                                                 prerelease_value: Optional[str]):
        hopla_version = HoplaVersion(pre_release=prerelease_value)
        result: str = hopla_version.semantic_version()
        assert re.fullmatch(pattern=semver_pattern, string=result)

    def test_semantic_version_ok(self, semver_pattern: Pattern):
        hopla_version = HoplaVersion(major_version=123,
                                     minor_version=200,
                                     patch_version=69,
                                     pre_release="beta")

        result: str = hopla_version.semantic_version()
        assert result == "123.200.69-beta"
        assert re.fullmatch(pattern=semver_pattern, string=result)

    @pytest.fixture
    def semver_pattern(self) -> Pattern:
        """Regex adopted from <https://semver.org/>"""
        return re.compile(
            r"(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)"
            r"(?:-("
            r"?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)"
            r"(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*"
            r"))?"
            r"(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?"
        )



#!/usr/bin/env python3
"""
The module with CLI code that handles the `hopla feed-all` command.
"""
import logging
import sys
from typing import Any, Callable, Dict, Iterator, List, NoReturn, Optional, Union

import click
from requests import Response

from hopla.hoplalib import hopla_option
from hopla.cli.groupcmds.get_user import HabiticaUser, HabiticaUserRequest
from hopla.hoplalib.throttling import RateLimitingAwareThrottler
from hopla.hoplalib.zoo.foodmodels import FoodStockpile, FoodStockpileBuilder
from hopla.hoplalib.zoo.petcontroller import FeedPostRequester
from hopla.hoplalib.zoo.zoomodels import Zoo, ZooBuilder
from hopla.hoplalib.zoo.zoofeed_algorithms import FeedAlgorithm, FeedPlan

log = logging.getLogger()


def __get_feed_plan_or_exit() -> Union[NoReturn, FeedPlan]:
    """Get the user and build the feed plan"""
    user: HabiticaUser = HabiticaUserRequest().request_user_data_or_exit()
    stockpile: FoodStockpile = FoodStockpileBuilder().user(user).build()
    zoo: Zoo = ZooBuilder(user).build()

    algorithm = FeedAlgorithm(zoo=zoo, stockpile=stockpile)
    return algorithm.make_plan()


def __confirm_with_user_or_abort(plan: FeedPlan) -> Optional[NoReturn]:
    """Ask the user to confirm the specified plan.

    :param plan: the feed plan to display
    :return: Don't return if the user wants to abort
    """

    prompt_msg = f"{plan.format_plan()}\nDo you want to proceed?"
    click.confirm(text=prompt_msg, abort=True)


def __feed_pets_without_confirmation(plan: FeedPlan):
    """Feed all the pets in the plan. Print the result to the terminal.

    Warning: this function does ask for confirmation.
    """
    # Feel free to refactor this such that we don't iterate over
    # the plan twice.
    feed_requests: List[Callable[[], Response]] = [
        FeedPostRequester.build_from(item).post_feed_request
        for item in plan
    ]
    throttler = RateLimitingAwareThrottler(feed_requests)
    feed_response_iter: Iterator[Response] = throttler.perform_and_yield_response()
    for item in plan:
        feed_response: Response = next(feed_response_iter)
        response_json: Dict[str, Any] = feed_response.json()
        if response_json["success"] is True:
            click.echo(response_json["message"])
        else:
            click.echo(f"Failed to feed {item.pet_name}\n"
                       f"{response_json['error']}: {response_json['message']}")


@click.command()
@hopla_option.no_interactive_option()
def feed_all(no_interactive: bool) -> None:
    """Feed all your pets.

    This command will first feed normal pets, then your quest pets, and
    finally all your pets that were hatched with magic hatching potions.

    Not this command will first show you the feed plan, and for safety,
    ask for your confirmation. Pets will only if you confirm this prompt.


    \b
    Examples
    --------
    # list the feeding plan, ask for confirmation, feed all pets if confirmed.
    $ hopla feed-all
    > Pet BearCub-Zombie will get 9 RottenMeat.
    > Pet Fox-CottonCandyBlue will get 9 CottonCandyBlue.
    > Do you want to proceed? [y/N]: yes


    \b
    # feed all the pets without asking for confirmation
    $ hopla feed-all --yes
    $ hopla feed-all --force

    \f
    :param no_interactive:
    """
    log.debug(f"hopla feed-all {no_interactive=}")
    plan: FeedPlan = __get_feed_plan_or_exit()
    if plan.isempty():
        click.echo(
            "The feed plan is empty. Reasons for this could be:\n"
            "1. There is insufficient food available to turn pets into mounts.\n"
            "2. You don't have any feedable pets."
        )
        sys.exit(0)

    if no_interactive is False:
        __confirm_with_user_or_abort(plan)

    __feed_pets_without_confirmation(plan)



"""
The module with CLI code that handles the `hopla get-user info` command.
"""

import logging

import click

from hopla.hoplalib.outputformatter import JsonFormatter
from hopla.cli.groupcmds.get_user import pass_user, HabiticaUser

log = logging.getLogger()

valid_info_names = click.Choice(["gems", "id", "notifications", "tags", "lastCron",
                                 "loginIncentives", "achievements",
                                 "guilds", "all"])


def info_alias_to_official_habitica_name(user_info_name: str) -> str:
    """Return the canonical habitica name"""
    if user_info_name in ["userid", "user-id"]:
        return "id"
    if user_info_name in ["gems", "gem"]:
        return "balance"
    return user_info_name


@click.command(context_settings={"token_normalize_func": info_alias_to_official_habitica_name})
@click.argument("user_info_name", type=valid_info_names, default="all")
@pass_user
def info(user: HabiticaUser,
         user_info_name: str) -> dict:
    """Return user information

    USER_INFO_NAME is the particular type of information that you want to return, by
    default "all".

    \b
    Examples:
    ---
    $ hopla get-user info                  # get all user info
    $ hopla get-user info gems             # get number of gems
    $ hopla get-user info id               # get user id
    $ hopla get-user info loginIncentives  # get number of times logged in

    \b
    # Examples with a jq filter:
    $ hopla get-user info | jq .items              # get all items of a user
    $ hopla get-user info | jq ".items.mounts"     # get all mounts
    $ hopla get-user info | jq ".inbox.messages"   # get personal messages

    \b
    # first get the notifications, then use jq to filter the notifications
    $ hopla get-user info notifications | jq '.[].type'

    \b
    # first get all tag objects and then only get the names of the tags
    $ hopla get-user info tags | jq '.[].name'

    \f
    [APIdocs](https://habitica.com/apidoc/#api-User-UserGet)

    :param user: HabiticaUser
    :param user_info_name:
    :return
    """
    log.debug(f"hopla get-user info {user_info_name=}")

    filtered_user = filter_on_user_info_name(user, user_info_name)

    user_str = JsonFormatter(filtered_user).format_with_double_quotes()
    click.echo(user_str)
    return filtered_user


def filter_on_user_info_name(user: HabiticaUser, user_info_name: str):
    """First time we filter the user.
       Returns whatever the result of filtering """
    if user_info_name == "gems":
        return user.get_gems()

    if user_info_name not in [None, "all"]:
        return user[user_info_name]

    return user.user_dict



#!/usr/bin/env python3
from typing import Dict, List
from unittest.mock import MagicMock, patch

import pytest
from click.testing import CliRunner, Result

from hopla.cli.hatch_all import hatch_all, to_pet_list
from hopla.hoplalib.hopla_option import NO_INTERACTION_OPTION_NAMES
from hopla.hoplalib.user.usermodels import HabiticaUser
from hopla.hoplalib.zoo.foodmodels import FeedStatus
from hopla.hoplalib.zoo.petmodels import Pet


class MockThrottler:
    def __init__(self, *, response1: str, response2: str):
        self.response1 = response1
        self.response2 = response2

    def release(self):
        yield lambda: self.response1
        yield lambda: self.response2


class MockOkHatchResponse:
    def __init__(self, msg):
        self.msg = msg

    def json(self):
        return {"success": True, "message": self.msg}


class TestHatchAllCliCommand:

    @patch("hopla.cli.hatch_all.HabiticaUserRequest.request_user_data_or_exit")
    def test_hatch_all_nothing_to_hatch_fail(self, mock_user_request: MagicMock):
        nothing_to_hatch_user = HabiticaUser({"items": {
            "pets": {"Wolf-Base": 5},
            "eggs": {"Wolf": 2},
            "hatchingPotions": {"Base": 1}
        }})
        mock_user_request.return_value = nothing_to_hatch_user

        runner = CliRunner()
        result: Result = runner.invoke(hatch_all)

        assert result.exit_code == 1
        assert result.stdout == (
            "The hatch plan is empty. Do you have enough eggs and hatching potions?\n"
            "Exiting\n"
        )

    @patch("hopla.cli.hatch_all.HabiticaUserRequest.request_user_data_or_exit")
    def test_hatch_all_something_to_hatch_user_denies_ok(self, mock_user_request: MagicMock):
        egg_name = "Wolf"
        potion1_name = "Base"
        potion2_name = "Fluorite"
        mock_user_request.return_value = HabiticaUser({"items": {
            "pets": {"Turtle-White": 15},
            "eggs": {egg_name: 2},
            "hatchingPotions": {potion1_name: 1, potion2_name: 1}
        }})

        runner = CliRunner()
        user_denies_input: str = "No"
        result: Result = runner.invoke(hatch_all, input=user_denies_input)

        expected_msg = (
            f"A {egg_name} egg will be hatched by a {potion1_name} potion.\n"
            f"A {egg_name} egg will be hatched by a {potion2_name} potion.\n"
            f"Do you wish to proceed? [y/N]: {user_denies_input}\n"
            "No eggs were hatched.\n"
        )
        # User denying is not an error. This is part of the normal flow. So we want a 0.
        assert result.exit_code == 0
        assert result.stdout.endswith(expected_msg)

    @patch("hopla.cli.hatch_all.RateLimitingAwareThrottler.perform_and_yield_response")
    @patch("hopla.cli.hatch_all.HabiticaUserRequest.request_user_data_or_exit")
    def test_hatch_all_something_to_hatch_user_confirms_ok(self,
                                                           mock_user_request: MagicMock,
                                                           mock_throttler: MagicMock):
        egg_name = "Wolf"
        potion1_name = "Base"
        potion2_name = "Fluorite"
        mock_user_request.return_value = HabiticaUser({"items": {
            "pets": {},
            "eggs": {egg_name: 2},
            "hatchingPotions": {potion1_name: 1, potion2_name: 1}
        }})

        msg1 = f"Successfully hatched a {egg_name}-{potion1_name}."
        msg2 = f"Successfully hatched a {egg_name}-{potion2_name}."
        response1 = MockOkHatchResponse(msg1)
        response2 = MockOkHatchResponse(msg2)

        mock_throttler.return_value = iter([response1, response2])

        runner = CliRunner()
        user_confirms_input: str = "yes"
        result: Result = runner.invoke(hatch_all, input=user_confirms_input)

        expected_msg = (
            f"A {egg_name} egg will be hatched by a {potion1_name} potion.\n"
            f"A {egg_name} egg will be hatched by a {potion2_name} potion.\n"
            f"Do you wish to proceed? [y/N]: {user_confirms_input}\n"
            f"{msg1}\n"
            f"{msg2}\n"
        )
        assert result.exit_code == 0
        assert result.stdout == expected_msg

    @pytest.mark.parametrize("force_option", NO_INTERACTION_OPTION_NAMES)
    @patch("hopla.cli.hatch_all.RateLimitingAwareThrottler.perform_and_yield_response")
    @patch("hopla.cli.hatch_all.HabiticaUserRequest.request_user_data_or_exit")
    def test_hatch_all_something_to_hatch_force_ok(self,
                                                   mock_user_request: MagicMock,
                                                   mock_throttler: MagicMock,
                                                   force_option: str):
        egg_name = "Wolf"
        potion1_name = "Base"
        potion2_name = "Fluorite"
        mock_user_request.return_value = HabiticaUser({"items": {
            "pets": {},
            "eggs": {egg_name: 2},
            "hatchingPotions": {potion1_name: 1, potion2_name: 1}
        }})

        msg1 = f"Successfully hatched a {egg_name}-{potion1_name}."
        msg2 = f"Successfully hatched a {egg_name}-{potion2_name}."
        response1 = MockOkHatchResponse(msg1)
        response2 = MockOkHatchResponse(msg2)

        mock_throttler.return_value = iter([response1, response2])

        runner = CliRunner()
        result: Result = runner.invoke(hatch_all, [force_option])

        assert result.exit_code == 0
        assert result.stdout == f"{msg1}\n{msg2}\n"


class TestToPetList:
    def test_to_pet_list_empty_dict_ok(self):
        result: List[Pet] = to_pet_list({})
        assert result == []

    def test_to_pet_list_ok(self):
        pet1, feed_status1 = "Wolf-Base", 5
        pet2, feed_status2 = "Axolotl-Red", 10
        pet3, feed_status3 = "Rat-White", -1
        pet4, feed_status4 = "Turtle-Golden", 0

        pets: Dict[str, int] = {
            pet1: feed_status1, pet2: feed_status2,
            pet3: feed_status3, pet4: feed_status4
        }

        result: List[Pet] = to_pet_list(pets)

        expected: List[Pet] = [
            Pet(pet1, feed_status=FeedStatus(feed_status1)),
            Pet(pet2, feed_status=FeedStatus(feed_status2)),
            Pet(pet3, feed_status=FeedStatus(feed_status3)),
            Pet(pet4, feed_status=FeedStatus(feed_status4))
        ]
        assert result == expected



#!/usr/bin/env python3
"""
Module with controllers for buying things from the Habitica API.
"""
from dataclasses import dataclass
from typing import NoReturn, Union

import requests

from hopla.hoplalib.requests_helper import get_data_or_exit
from hopla.hoplalib.http import HabiticaRequest, UrlBuilder


@dataclass
class BuyEnchantedArmoireRequest(HabiticaRequest):
    """HabiticaRequest that buys from the enchanted armoire."""

    @property
    def url(self) -> str:
        """The habitica API buy url."""
        return UrlBuilder(path_extension="/user/buy-armoire").url

    def post_buy_request(self) -> requests.Response:
        """POST a buy request to the habitica API."""
        return requests.post(
            url=self.url,
            headers=self.default_headers,
            timeout=HabiticaRequest.TIMEOUT
        )

    def post_buy_request_get_data_or_exit(self) -> Union[dict, NoReturn]:
        """POST a buy request and return the result, exit if the request failed.

        :return: If successful, the armoire content.
        """
        response: requests.Response = self.post_buy_request()
        # By default, we get way too much JSON info, so filter on "armoire".
        return get_data_or_exit(response)["armoire"]



"""
The module with CLI code that handles the `hopla add` group command.
"""
import logging

import click

log = logging.getLogger()


@click.group()
def add():
    """GROUP for adding things to Habitica."""



"""
The module with CLI code that handles the `hopla authenticate` command.
"""
import logging
from uuid import UUID
from dataclasses import dataclass

import click

from hopla.hoplalib.authorization import AuthorizationHandler

log = logging.getLogger()


@click.command()
def authenticate():
    """Authorize yourself to access the Habitica.com API.

    hopla authenticate allows you to interactively provide access credentials.
    """
    log.debug("hopla authenticate")

    hopla_user_credentials: HoplaUserCredentials = request_user_for_credentials()

    AuthorizationHandler().set_hopla_credentials(
        user_id=hopla_user_credentials.user_id,
        api_token=hopla_user_credentials.api_token,
        overwrite=True
    )


@dataclass(frozen=True)
class HoplaUserCredentials:
    """Class representing the user's credentials"""
    user_id: UUID
    api_token: UUID


def request_user_for_credentials() -> HoplaUserCredentials:
    """Request the user for the user id and api token"""
    click.echo("Please enter your credentials.")
    click.echo("You can find them over at <https://habitica.com/user/settings/api>.")
    click.echo("They have the following format: 'c0ffee69-dada-feed-abb1-5ca1ab1ed004'.")
    click.echo("The user id can be found under 'User ID'.")
    user_id = click.prompt(text="Please paste your user ID here (press Ctrl-C to abort)",
                           type=click.UUID)
    api_token = click.prompt(text="Please paste your user API Token here (input remains hidden)",
                             type=click.UUID, hide_input=True)
    return HoplaUserCredentials(user_id=user_id, api_token=api_token)



"""
The module with CLI code that handles the `hopla get-user stats` command.
"""

import logging
from typing import Any

import click

from hopla.hoplalib.outputformatter import JsonFormatter
from hopla.cli.groupcmds.get_user import pass_user, HabiticaUser

log = logging.getLogger()

valid_stat_names = click.Choice(["hp", "mp", "exp", "gp", "lvl", "class",
                                 "maxMP",
                                 "int", "str", "per", "con",
                                 "all"])


def stat_alias_to_official_habitica_name(stat_name: str) -> str:  # noqa: C901 (Mccabe too high)
    """Turn typos and similar namings into stats recognized by the habitica API.

    :param stat_name:
    :return:
    """
    # pylint: disable=too-many-return-statements,too-many-branches
    if stat_name in ["mana", "mana-points", "manapoints"]:
        return "mp"
    if stat_name in ["maxMp", "maxmp"]:
        return "maxMP"
    if stat_name in ["health", "healthpoints"]:
        return "hp"
    if stat_name in ["xp", "experience"]:
        return "exp"
    if stat_name in ["gold"]:
        return "gp"
    if stat_name in ["level"]:
        return "lvl"
    if stat_name in ["intelligence"]:
        return "int"
    if stat_name in ["strength"]:
        return "str"
    if stat_name in ["perception"]:
        return "per"
    if stat_name in ["constitution"]:
        return "con"

    return stat_name


@click.command(
    context_settings={"token_normalize_func": stat_alias_to_official_habitica_name}
)
@click.argument("stat_name", type=valid_stat_names, default="all")
@pass_user
def stats(user: HabiticaUser, stat_name: str) -> Any:
    """Get the stats of a user

    \b
    Examples
    ---
    # get all user stats
    $ hopla get-user stats
    $ hopla get-user stats all

    \b
    # get-user mana, health, level
    $ hopla get-user stats mp
    $ hopla get-user stats hp
    $ hopla get-user stats lvl
    """
    log.debug(f"hopla get-user stats {stat_name=}")

    stats_data = user.get_stats()
    if stat_name == "all":
        requested_user_data = stats_data
    else:
        requested_user_data = stats_data[stat_name]
    click.echo(JsonFormatter(requested_user_data).format_with_double_quotes())
    return requested_user_data



#!/usr/bin/env python3
import pytest

from hopla.hoplalib.outputformatter import JsonFormatter


class TestJsonFormatter:
    def test___init__(self, json_from_habitica: dict):
        formatter = JsonFormatter(json_from_habitica)

        assert formatter.json_as_dict == json_from_habitica

    def test_format_with_double_quotes(self, json_from_habitica):
        indent = 2
        formatter = JsonFormatter(json_from_habitica)

        result: str = formatter.format_with_double_quotes(indent=indent)

        # I'm using the 'sp' variable name here, because it is also 2 chars long.
        sp: str = ' ' * indent
        assert result == (
            '{\n'
            f'{sp}"success": true,\n'
            f'{sp}"data": -1,\n'
            f'{sp}"message": "You have tamed Shade Velociraptor, let\'s go for a ride!"\n'
            '}'
        )

    @pytest.fixture
    def json_from_habitica(self) -> dict:
        return {
            'success': True, 'data': -1,
            'message': 'You have tamed Shade Velociraptor, let\'s go for a ride!'
        }



#!/usr/bin/env python3
from typing import List


class HabiticaClassData:
    """A class with data about habitica classes.

    @see: hopla api content | jq .classes
    """
    class_names: List[str] = ["warrior", "rogue", "healer", "wizard"]



#!/usr/bin/env python3
from typing import List, Optional
from unittest.mock import _Call, call, MagicMock, patch

import pytest
from click.testing import CliRunner, Result

from hopla.cli.cast import cast, times_until_out_of_mana
from hopla.hoplalib.cast.spellmodel import Spell


class MockCastResponse:
    def __init__(self, success: bool,
                 user_mp: float,
                 message: Optional[str],
                 error: Optional[str],
                 status_code: int):
        self.success = success
        self.user_mp = user_mp
        self.message = message
        self.error = error
        self.status_code = status_code

    def json(self) -> dict:
        if self.success is True:
            return {
                "success": self.success,
                "data": {"user": {"stats": {"mp": self.user_mp}}}
            }

        return {
            "success": self.success,
            "message": self.message,
            "error": self.error
        }


class MockCastRequest:

    def __init__(self, success: bool, remaining_user_mana: float = 0.0,
                 message: Optional[str] = None, error: Optional[str] = None,
                 status_code: int = 200):
        self.success = success
        self.user_mp = remaining_user_mana
        self.message = message
        self.error = error
        self.status_code = status_code

    def post_spell(self):
        return MockCastResponse(success=self.success, user_mp=self.user_mp,
                                message=self.message, error=self.error,
                                status_code=self.status_code)


class TestCastCliCommand:
    @patch("hopla.cli.cast.PostCastRequest")
    def test_cast_once_ok(self, mock_cast_request: MagicMock):
        mana_left = 11.1
        mock_cast_request.return_value = MockCastRequest(success=True,
                                                         remaining_user_mana=mana_left)
        spell_name = "intimidate"

        runner = CliRunner()
        result: Result = runner.invoke(cast, [spell_name])

        expected_spell = Spell(spell_name)
        mock_cast_request.assert_called_with(spell=expected_spell)

        expected_output = f"{spell_name} casted successfully: {mana_left:.1f} mana left.\n"
        assert result.stdout == expected_output

        assert result.exit_code == 0

    @patch("hopla.cli.cast.PostCastRequest")
    def test_cast_once_fail(self, mock_cast_request: MagicMock):
        err_msg = "Not enough mana."
        error_name = "NotAuthorized"
        mock_cast_request.return_value = MockCastRequest(
            success=False, message=err_msg, error=error_name,
            status_code=400
            # Yes, I know..., but according to the docs 400 is returned with NotAuthorized:
            # https://habitica.com/apidoc/#api-User-UserCast
        )

        spell_name = "protectAura"

        runner = CliRunner()
        result: Result = runner.invoke(cast, [spell_name])

        expected_spell = Spell(spell_name)
        mock_cast_request.assert_called_with(spell=expected_spell)

        assert '"success": false' in result.stdout
        assert f'"message": "{err_msg}"' in result.stdout
        assert f'"error": "{error_name}"' in result.stdout

        assert result.exit_code == 1

    @patch("hopla.cli.cast.PostCastRequest")
    def test_cast_until_out_of_mana_once_ok(self, mock_cast_request: MagicMock):
        mana_left = 1.1
        spell_name = "toolsOfTrade"
        mock_cast_request.return_value = MockCastRequest(success=True,
                                                         remaining_user_mana=mana_left)

        runner = CliRunner()
        result: Result = runner.invoke(cast, [spell_name, "--until-out-of-mana"])

        expected_spell = Spell(spell_name)
        assert mock_cast_request.call_args_list == [call(spell=expected_spell)]

        assert result.exit_code == 0

    @pytest.mark.parametrize("spell_name,mana_left,num_expected_calls", [
        ("toolsOfTrade", 26.1, 2),
        ("stealth", 12.0, 1),
        ("earth", 69., 2),
        ("earth", 71.1, 3)
    ])
    @patch("hopla.cli.cast.PostCastRequest")
    def test_cast_until_out_of_mana_multiple_times_ok(self, mock_cast_request: MagicMock,
                                                      spell_name: str,
                                                      mana_left: float,
                                                      num_expected_calls: int):
        mock_cast_request.return_value = MockCastRequest(success=True,
                                                         remaining_user_mana=mana_left)

        runner = CliRunner()
        result: Result = runner.invoke(cast, [spell_name, "-u"])

        expected_call_list: List[_Call] = [
            call(spell=Spell(spell_name)) for _ in range(num_expected_calls)
        ]
        assert mock_cast_request.call_args_list == expected_call_list

        assert result.exit_code == 0


class TestCastHelpers:
    @pytest.mark.parametrize("spell_name,remaining_mana,expected_times", [
        ("toolsOfTrade", 97., 3),
        ("toolsOfTrade", 101., 4),
        ("healAll", 101., 4),
        ("earth", 69., 1),
        ("earth", 71.2, 2),
        ("stealth", 1.2, 0),
        ("defensiveStance", 50.0, 2)
    ])
    def test_times_until_out_of_mana(self,
                                     spell_name: str,
                                     remaining_mana: float,
                                     expected_times: int):
        spell = Spell(name=spell_name)
        result_times: int = times_until_out_of_mana(spell, remaining_mana=remaining_mana)
        assert result_times == expected_times



#!/usr/bin/env python3
import logging
import re
from re import Pattern

import pytest
from click.testing import CliRunner, Result

from hopla import hopla
from hopla.cli.version import version
from hopla.kickstart import setup_logging


class TestVersionCliCommand:
    def test_version(self, semver_pattern: Pattern):
        runner = CliRunner()
        result: Result = runner.invoke(version)

        assert re.fullmatch(pattern=semver_pattern, string=result.stdout)
        assert result.exit_code == 0

    def test_version_similar_to_hopla_version_option(self, semver_pattern: Pattern):
        runner = CliRunner()
        result_cmd: Result = runner.invoke(version)
        result_option: Result = runner.invoke(hopla, ["--version"])

        # hopla --version prefixes 'hopla, version'
        assert result_option.stdout == f"hopla, version {result_cmd.stdout}"

    @pytest.fixture
    def semver_pattern(self) -> Pattern:
        """Regex adopted from <https://semver.org/>"""
        return re.compile(
            r"(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)"
            r"(?:-("
            r"?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)"
            r"(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*"
            r"))?"
            r"(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?"
            "\n"  # added the \n here
        )


class TestHoplaCommand:
    def test_hopla_base_cmd_gives_help(self):
        runner = CliRunner()
        result_hopla: Result = runner.invoke(hopla)
        result_h: Result = runner.invoke(hopla, ["-h"])
        result_help: Result = runner.invoke(hopla, ["--help"])

        assert result_hopla.exit_code == 0
        assert result_help.exit_code == 0
        assert result_hopla.stdout == result_help.stdout
        assert result_h.stdout == result_help.stdout


class TestSetupLogging:
    def test_setup_logging(self):
        result_logger: logging.Logger = setup_logging()

        assert result_logger.name == "hopla"
